'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx, os
from taskcoachlib.persistence import BackupManifest
from taskcoachlib.i18n import _
from taskcoachlib import render


class BackupManagerDialog(wx.Dialog):
    def __init__(self, parent, settings, selectedFile=None):
        super(BackupManagerDialog, self).__init__(parent, wx.ID_ANY, style=wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER)

        container = wx.Panel(self)
        self.__files = wx.ListCtrl(container, wx.ID_ANY, style=wx.LC_REPORT|wx.LC_SINGLE_SEL)
        self.__files.InsertColumn(0, _('File'))
        self.__files.InsertColumn(1, _('Full path'))
        self.__backups = wx.ListCtrl(container, wx.ID_ANY, style=wx.LC_REPORT|wx.LC_SINGLE_SEL)
        self.__backups.InsertColumn(0, _('Date'))
        self.__backups.Enable(False)
        self.__btnRestore = wx.Button(container, wx.ID_ANY, _('Restore'))
        self.__btnRestore.Enable(False)
        self.__filename = selectedFile
        self.__selection = (None, None)

        vsz = wx.BoxSizer(wx.VERTICAL)
        hsz = wx.BoxSizer(wx.HORIZONTAL)
        hsz.Add(self.__files, 1, wx.EXPAND|wx.ALL, 3)
        hsz.Add(self.__backups, 1, wx.EXPAND|wx.ALL, 3)
        hsz.Add(self.__btnRestore, 0, wx.ALIGN_TOP|wx.ALL, 3)
        vsz.Add(hsz, 1, wx.EXPAND)

        container.SetSizer(vsz)

        vsz = wx.BoxSizer(wx.VERTICAL)
        vsz.Add(container, 1, wx.EXPAND)
        btn = wx.Button(self, wx.ID_ANY, _('Close'))
        vsz.Add(btn, 0, wx.ALL|wx.ALIGN_RIGHT, 3)
        self.SetSizer(vsz)

        self.__manifest = BackupManifest(settings)
        self.__filenames = self.__manifest.listFiles()
        selection = None
        for filename in self.__filenames:
            item = self.__files.InsertStringItem(self.__files.GetItemCount(), os.path.split(filename)[-1])
            self.__files.SetStringItem(item, 1, filename)
            if filename == selectedFile:
                selection = item

        self.SetSize(wx.Size(600, 400))
        self.CentreOnParent()

        wx.EVT_BUTTON(btn, wx.ID_ANY, self.DoClose)
        wx.EVT_LIST_ITEM_SELECTED(self.__files, wx.ID_ANY, self._OnSelectFile)
        wx.EVT_LIST_ITEM_DESELECTED(self.__files, wx.ID_ANY, self._OnDeselectFile)
        wx.EVT_LIST_ITEM_SELECTED(self.__backups, wx.ID_ANY, self._OnSelectBackup)
        wx.EVT_LIST_ITEM_DESELECTED(self.__backups, wx.ID_ANY, self._OnDeselectBackup)
        wx.EVT_BUTTON(self.__btnRestore, wx.ID_ANY, self._OnRestore)

        if selection is not None:
            self.__files.SetItemState(selection, wx.LIST_STATE_FOCUSED|wx.LIST_STATE_SELECTED, wx.LIST_STATE_FOCUSED|wx.LIST_STATE_SELECTED)
        self.__files.SetColumnWidth(0, -1)
        self.__files.SetColumnWidth(1, -1)

    def restoredFilename(self):
        return self.__filename

    def DoClose(self, event):
        self.EndModal(wx.ID_CANCEL)

    def _OnSelectFile(self, event):
        self.__backups.DeleteAllItems()
        for index, dateTime in enumerate(self.__manifest.listBackups(self.__filenames[event.GetIndex()])):
            self.__backups.InsertStringItem(index, render.dateTime(dateTime, humanReadable=True))
        self.__backups.SetColumnWidth(0, -1)
        self.__backups.Enable(True)
        self.__selection = (self.__filenames[event.GetIndex()], None)

    def _OnDeselectFile(self, event):
        self.__btnRestore.Enable(False)
        self.__backups.DeleteAllItems()
        self.__backups.Enable(False)
        self.__selection = (None, None)

    def _OnSelectBackup(self, event):
        self.__btnRestore.Enable(True)
        filename = self.__selection[0]
        self.__selection = (filename, self.__manifest.listBackups(filename)[event.GetIndex()])

    def _OnDeselectBackup(self, event):
        self.__btnRestore.Enable(False)
        self.__selection = (self.__selection[0], None)

    def _OnRestore(self, event):
        filename, dateTime = self.__selection
        dlg = wx.FileDialog(self, _('Choose the restoration destination'),
                            defaultDir=os.path.dirname(filename),
                            defaultFile=os.path.split(filename)[-1],
                            wildcard='*.tsk', style=wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)
        try:
            if dlg.ShowModal() == wx.ID_OK:
                self.__filename = dlg.GetPath()
                self.__manifest.restoreFile(filename, dateTime, self.__filename)
                self.EndModal(wx.ID_OK)
        finally:
            dlg.Destroy()



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import codecs, os
from taskcoachlib.thirdparty.pubsub import pub
import todotxt


class AutoImporterExporter(object):
    ''' AutoImporterExporter observes task files. If a task file is saved, 
        either by the user or automatically (when autosave is on) and auto 
        import and/or export is on, AutoImporterExporter imports and/or exports 
        the task file. '''
        
    def __init__(self, settings):
        super(AutoImporterExporter, self).__init__()
        self.__settings = settings
        pub.subscribe(self.onTaskFileAboutToBeSaved, 'taskfile.aboutToSave')
        pub.subscribe(self.onTaskFileJustRead, 'taskfile.justRead')
        
    def onTaskFileJustRead(self, taskFile):
        ''' After a task file has been read and if auto import is on, 
            import it. '''
        self.importFiles(taskFile)
            
    def onTaskFileAboutToBeSaved(self, taskFile):
        ''' When a task file is about to be saved and auto import and/or 
            export is on, import and/or export it. '''
        self.importFiles(taskFile)
        self.exportFiles(taskFile)
        
    def importFiles(self, taskFile):
        importFormats = self.__settings.getlist('file', 'autoimport')
        for importFormat in importFormats:
            if importFormat == 'Todo.txt':
                self.importTodoTxt(taskFile)
                
    def exportFiles(self, taskFile):
        exportFormats = self.__settings.getlist('file', 'autoexport')
        for exportFormat in exportFormats:
            if exportFormat == 'Todo.txt':
                self.exportTodoTxt(taskFile)
    
    @classmethod            
    def importTodoTxt(cls, taskFile):
        filename = cls.todoTxtFilename(taskFile)
        if os.path.exists(filename):
            todotxt.TodoTxtReader(taskFile.tasks(), taskFile.categories()).read(filename)

    @classmethod
    def exportTodoTxt(cls, taskFile):
        filename = cls.todoTxtFilename(taskFile)
        with codecs.open(filename, 'w', 'utf-8') as todoFile:
            todotxt.TodoTxtWriter(todoFile, filename).writeTasks(taskFile.tasks())
    
    @staticmethod   
    def todoTxtFilename(taskFile):
        return taskFile.filename()[:-len('tsk')] + 'txt'



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import patterns, meta, notify
from taskcoachlib.domain import date, task
from taskcoachlib.gui.dialog import reminder, editor
from taskcoachlib.i18n import _
from taskcoachlib.thirdparty.pubsub import pub
import wx


class ReminderController(object):
    def __init__(self, mainWindow, taskList, effortList, settings):
        super(ReminderController, self).__init__()
        pub.subscribe(self.onSetReminder, task.Task.reminderChangedEventType())
        patterns.Publisher().registerObserver(self.onAddTask,
            eventType=taskList.addItemEventType(),
            eventSource=taskList)
        patterns.Publisher().registerObserver(self.onRemoveTask,
            eventType=taskList.removeItemEventType(),
            eventSource=taskList)
        self.__tasksWithReminders = {}  # {task: reminderDateTime}
        self.__mainWindow = mainWindow
        self.__mainWindowWasHidden = False
        self.__registerRemindersForTasks(taskList)
        self.settings = settings
        self.taskList = taskList
        self.effortList = effortList

    def onAddTask(self, event):
        self.__registerRemindersForTasks(event.values())
                
    def onRemoveTask(self, event):
        self.__removeRemindersForTasks(event.values())
                
    def onSetReminder(self, newValue, sender):  # pylint: disable=W0613
        self.__removeRemindersForTasks([sender])
        self.__registerRemindersForTasks([sender])
        
    def onReminder(self):
        self.showReminderMessages(date.DateTime.now())
        
    def showReminderMessages(self, now):
        now += date.TimeDelta(seconds=5)  # Be sure not to miss reminders 
        requestUserAttention = False
        for taskWithReminder in self.__tasksWithReminders.copy():
            if taskWithReminder.reminder() <= now:
                requestUserAttention = True
                self.showReminderMessage(taskWithReminder)
        if requestUserAttention:
            self.requestUserAttention()        
        
    def showReminderMessage(self, taskWithReminder, 
                            ReminderDialog=reminder.ReminderDialog):
        if self.__useOwnReminderDialog():
            self.__showReminderDialog(taskWithReminder, ReminderDialog)
            self.__removeReminder(taskWithReminder)
        else:
            self.__showReminderViaNotifier(taskWithReminder)
            self.__removeReminder(taskWithReminder)
            self.__snooze(taskWithReminder)
            
    def __useOwnReminderDialog(self):
        notifier = self.settings.get('feature', 'notifier')
        return notifier == 'Task Coach' or notify.AbstractNotifier.get(notifier) is None
        
    def __showReminderDialog(self, taskWithReminder, ReminderDialog):
        # If the dialog has self.__mainWindow as parent, it steals the focus when
        # returning to Task Coach through Alt+Tab; we don't want that for
        # reminders.
        reminderDialog = ReminderDialog(taskWithReminder, self.taskList, 
                                        self.effortList, self.settings, None)
        reminderDialog.Bind(wx.EVT_CLOSE, self.onCloseReminderDialog)
        reminderDialog.Show()

    def __showReminderViaNotifier(self, taskWithReminder):
        notifier = notify.AbstractNotifier.get(self.settings.get('feature', 'notifier'))
        notifier.notify(_('%s Reminder') % meta.name, taskWithReminder.subject(),
                        wx.ArtProvider.GetBitmap('taskcoach', size=wx.Size(32, 32)),
                        windowId=self.__mainWindow.GetHandle())
        
    def __snooze(self, taskWithReminder):
        minutesToSnooze = self.settings.getint('view', 'defaultsnoozetime')
        taskWithReminder.snoozeReminder(date.TimeDelta(minutes=minutesToSnooze))
        
    def onCloseReminderDialog(self, event, show=True):
        event.Skip()
        dialog = event.EventObject
        taskWithReminder = dialog.task
        if not dialog.ignoreSnoozeOption:
            snoozeOptions = dialog.snoozeOptions
            snoozeTimeDelta = snoozeOptions.GetClientData(snoozeOptions.Selection)
            taskWithReminder.snoozeReminder(snoozeTimeDelta)  # Note that this is not undoable
            # Undoing the snoozing makes little sense, because it would set the 
            # reminder back to its original date-time, which is now in the past.
        if dialog.openTaskAfterClose:
            editTask = editor.TaskEditor(self.__mainWindow, [taskWithReminder], 
                self.settings, self.taskList, self.__mainWindow.taskFile, 
                bitmap='edit')
            editTask.Show(show)
        else:
            editTask = None
        dialog.Destroy()
        if self.__mainWindowWasHidden:
            self.__mainWindow.Hide()
        return editTask  # For unit testing purposes

    def requestUserAttention(self):
        notifier = self.settings.get('feature', 'notifier')
        if notifier != 'Task Coach' and notify.AbstractNotifier.get(notifier) is not None:
            # When using Growl/Snarl, this is not necessary. Even when not using Growl, it's
            # annoying as hell. Anyway.
            return
        self.__mainWindowWasHidden = not self.__mainWindow.IsShown()
        if self.__mainWindowWasHidden:
            self.__mainWindow.Show()
        if not self.__mainWindow.IsActive():
            self.__mainWindow.RequestUserAttention()
            
    def __registerRemindersForTasks(self, tasks):
        for eachTask in tasks:
            if eachTask.reminder() and eachTask.reminder() < date.DateTime():
                self.__registerReminder(eachTask)

    def __removeRemindersForTasks(self, tasks):
        for eachTask in tasks:
            if eachTask in self.__tasksWithReminders:
                self.__removeReminder(eachTask)

    def __registerReminder(self, taskWithReminder):
        reminderDateTime = taskWithReminder.reminder()
        now = date.DateTime.now()
        if reminderDateTime < now:
            reminderDateTime = now + date.TimeDelta(seconds=10)
        self.__tasksWithReminders[taskWithReminder] = date.Scheduler().schedule(self.onReminder, 
                                                                                reminderDateTime)
            
    def __removeReminder(self, taskWithReminder):
        job = self.__tasksWithReminders[taskWithReminder]
        date.Scheduler().unschedule(job)
        del self.__tasksWithReminders[taskWithReminder]



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test, wx
from taskcoachlib import patterns
from taskcoachlib.domain import note, date, base


class NoteTest(test.TestCase):
    def setUp(self):
        self.note = note.Note()
        self.child = note.Note()
        self.events = []
        
    def onEvent(self, event):
        self.events.append(event)
        
    def testDefaultSubject(self):
        self.assertEqual('', self.note.subject())
        
    def testGivenSubject(self):
        aNote = note.Note(subject='Note')
        self.assertEqual('Note', aNote.subject())
        
    def testSetSubject(self):
        self.note.setSubject('Note')
        self.assertEqual('Note', self.note.subject())
        
    def testSubjectChangeNotification(self):
        patterns.Publisher().registerObserver(self.onEvent, 
            self.note.subjectChangedEventType())
        self.note.setSubject('Note')
        self.assertEqual(patterns.Event(self.note.subjectChangedEventType(), 
            self.note, 'Note'), self.events[0])
        
    def testDefaultDescription(self):
        self.assertEqual('', self.note.description())
        
    def testGivenDescription(self):
        aNote = note.Note(description='Description')
        self.assertEqual('Description', aNote.description())
        
    def testSetDescription(self):
        self.note.setDescription('Description')
        self.assertEqual('Description', self.note.description())
        
    def testDescriptionChangeNotification(self):
        patterns.Publisher().registerObserver(self.onEvent, 
            self.note.descriptionChangedEventType())
        self.note.setDescription('Description')
        self.assertEqual(patterns.Event(self.note.descriptionChangedEventType(),
            self.note, 'Description'), self.events[0])
        
    def testAddChild(self):
        self.note.addChild(self.child)
        self.assertEqual([self.child], self.note.children())

    def testRemoveChild(self):
        self.note.addChild(self.child)
        self.note.removeChild(self.child)
        self.assertEqual([], self.note.children())
        
    def testAddChildNotification(self):
        patterns.Publisher().registerObserver(self.onEvent, 
            note.Note.addChildEventType())
        self.note.addChild(self.child)
        self.assertEqual(patterns.Event(note.Note.addChildEventType(), 
            self.note, self.child), self.events[0])
        
    def testRemoveChildNotification(self):
        patterns.Publisher().registerObserver(self.onEvent, 
            note.Note.removeChildEventType())
        self.note.addChild(self.child)
        self.note.removeChild(self.child)
        self.assertEqual(patterns.Event(note.Note.removeChildEventType(), 
            self.note, self.child), self.events[0])
        
    def testNewChild(self):
        child = self.note.newChild(subject='child')
        self.assertEqual('child', child.subject())  # pylint: disable=E1101
        
    def testGetState(self):
        self.assertEqual(dict(id=self.note.id(), subject='', description='', 
            parent=None, categories=set(), attachments=[], 
            children=self.note.children(), status=self.note.getStatus(), 
            fgColor=None, bgColor=None, font=None, icon='', selectedIcon='', 
            creationDateTime=self.note.creationDateTime(),
            modificationDateTime=self.note.modificationDateTime(),
            ordering=self.note.ordering()),
            self.note.__getstate__())
        
    def testSetState(self):
        self.note.__setstate__(dict(id='id', subject='new', description='new', 
            parent=None, children=[], status=42, attachments=[], categories=[],
            fgColor=(1, 1, 1, 1), bgColor=(0, 0, 0, 255), font=wx.SWISS_FONT,
            icon='icon', selectedIcon='selected', creationDateTime=date.Now(),
            modificationDateTime=date.Now(), ordering=42L))
        self.assertEqual('new', self.note.description())


class NoteOwnerUnderTest(note.NoteOwner, base.Object):
    pass

        
class NoteOwnerTest(test.TestCase):
    def setUp(self):
        self.note = note.Note(subject='Note')
        self.noteOwner = NoteOwnerUnderTest()
        self.events = []
        
    def onEvent(self, event):
        self.events.append(event)
    
    # pylint: disable=E1101
        
    def registerObserver(self):  # pylint: disable=W0221
        patterns.Publisher().registerObserver(self.onEvent,
            NoteOwnerUnderTest.notesChangedEventType())
        
    def testAddNote(self):
        self.noteOwner.addNote(self.note)
        self.assertEqual([self.note], self.noteOwner.notes())
        
    def testAddNotes(self):
        self.noteOwner.addNotes(self.note)
        self.assertEqual([self.note], self.noteOwner.notes())

    def testAddNoteNotification(self):
        self.registerObserver()
        self.noteOwner.addNote(self.note)
        self.assertEqual(patterns.Event( \
            NoteOwnerUnderTest.notesChangedEventType(), self.noteOwner, self.note), 
            self.events[0])

    def testRemoveNote(self):
        self.noteOwner.addNote(self.note)
        self.noteOwner.removeNote(self.note)
        self.failIf(self.noteOwner.notes())
        
    def testRemoveNotes(self):
        self.noteOwner.addNote(self.note)
        self.noteOwner.removeNotes(self.note)
        self.failIf(self.noteOwner.notes())
        
    def testRemoveNoteNotification(self):
        self.noteOwner.addNote(self.note)
        self.registerObserver()
        self.noteOwner.removeNote(self.note)
        self.assertEqual([patterns.Event( \
            NoteOwnerUnderTest.notesChangedEventType(), self.noteOwner, self.note)], 
            self.events)
        
    def testGetState(self):
        self.noteOwner.addNote(self.note)
        self.assertEqual([self.note], self.noteOwner.__getstate__()['notes'])

    def testSetState(self):
        self.noteOwner.addNote(self.note)
        state = self.noteOwner.__getstate__()
        self.noteOwner.removeNote(self.note)
        self.noteOwner.__setstate__(state)
        self.assertEqual([self.note], self.noteOwner.notes())
        
    def testSetState_CausesNotification(self):
        self.noteOwner.addNote(self.note)
        state = self.noteOwner.__getstate__()
        self.noteOwner.removeNote(self.note)
        self.registerObserver()
        self.noteOwner.__setstate__(state)
        self.assertEqual(patterns.Event( \
            NoteOwnerUnderTest.notesChangedEventType(), self.noteOwner, self.note), 
            self.events[0])
            
    def testInitializeNotesViaConstructor(self):
        noteOwner = NoteOwnerUnderTest(notes=[self.note])
        self.assertEqual([self.note], noteOwner.notes())

    def testCopy(self):
        self.noteOwner.addNote(self.note)
        copy = NoteOwnerUnderTest(**self.noteOwner.__getcopystate__())
        self.assertNotEqual(copy.notes()[0].id(), self.note.id())
        self.assertEqual(copy.notes()[0].subject(), self.note.subject())

    def testCopy_NoteOwnerWithNoteWithSubNote(self):
        child = note.Note(subject='child')
        self.note.addChild(child)
        self.noteOwner.addNote(self.note)
        copy = NoteOwnerUnderTest(**self.noteOwner.__getcopystate__())
        childCopy = copy.notes()[0].children()[0]
        self.assertNotEqual(childCopy.id(), child.id())
        self.assertEqual(childCopy.subject(), child.subject())



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib import gui, config, widgets, persistence
from taskcoachlib.domain import task, date
from taskcoachlib.thirdparty import hypertreelist


class AuiManagedFrameWithDynamicCenterPane(widgets.AuiManagedFrameWithDynamicCenterPane):
    def AddBalloonTip(self, *args, **kwargs):
        pass


class Window(AuiManagedFrameWithDynamicCenterPane):
    def addPane(self, viewer, title, name='name', floating=False):
        super(Window, self).addPane(viewer, title, name, floating)


class ViewerTest(test.wxTestCase):
    def setUp(self):
        super(ViewerTest, self).setUp()
        self.settings = config.Settings(load=False)
        self.taskFile = persistence.TaskFile()
        self.task = task.Task('task')
        self.taskFile.tasks().append(self.task)
        self.window = Window(self.frame)
        self.viewerContainer = gui.viewer.ViewerContainer(self.window, 
            self.settings)
        self.viewer = self.createViewer()
        self.viewerContainer.addViewer(self.viewer)

    def tearDown(self):
        super(ViewerTest, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()

    def createViewer(self):
        return gui.viewer.TaskViewer(self.window, self.taskFile,
            self.settings)

    def testSelectAllViaWidget(self):
        self.viewer.widget.select_all()
        self.viewer.updateSelection()
        self.assertEqual([self.task], self.viewer.curselection())
        
    def testSelectAllViaWidgetWithMultipleItems(self):
        self.taskFile.tasks().append(task.Task('second'))
        self.viewer.widget.select_all()
        self.viewer.updateSelection()
        self.assertEqual(2, len(self.viewer.curselection()))
        
    def testSelectAll(self):
        self.viewer.select_all()
        self.viewer.endOfSelectAll()
        self.assertEqual([self.task], self.viewer.curselection())
        
    def testSelectAllWithMultipleItems(self):
        self.taskFile.tasks().append(task.Task('second'))
        self.viewer.select_all()
        self.viewer.endOfSelectAll()
        self.assertEqual(2, len(self.viewer.curselection()))
        
    def testSelectNextItemAfterDeletingSelection(self):
        secondTask = task.Task('second')
        self.taskFile.tasks().append(secondTask)
        self.viewer.select([self.task])
        self.taskFile.tasks().remove(self.task)
        self.assertEqual([secondTask], self.viewer.curselection())

    def testSelectParentAfterDeletingSelectedChild(self):
        secondTask = task.Task('second')
        self.taskFile.tasks().append(secondTask)
        child = task.Task('child')
        self.task.addChild(child)
        child.setParent(self.task)
        self.taskFile.tasks().append(child)
        self.viewer.select([child])
        self.taskFile.tasks().remove(child)
        self.assertEqual([self.task], self.viewer.curselection())
        
    def testDontChangeSelectionAfterDeletingAnItemThatIsNotSelected(self):
        secondTask = task.Task('second')
        self.taskFile.tasks().append(secondTask)
        child = task.Task('child')
        self.task.addChild(child)
        child.setParent(self.task)
        self.taskFile.tasks().append(child)
        self.viewer.select([secondTask])
        self.taskFile.tasks().remove(child)
        self.assertEqual([secondTask], self.viewer.curselection())
          
    def testFirstViewerInstanceSettingsSection(self):
        self.assertEqual(self.viewer.__class__.__name__.lower(), 
                         self.viewer.settingsSection())
        
    def testSecondViewerInstanceHasAnotherSettingsSection(self):
        viewer2 = self.createViewer()
        self.assertEqual(self.viewer.settingsSection() + '1', 
                         viewer2.settingsSection())
        
    def testTitle(self):
        self.assertEqual(self.viewer.defaultTitle, self.viewer.title())
        
    def testSetTitle(self):
        self.viewer.setTitle('New title')
        self.assertEqual('New title', self.viewer.title())
        
    def testSetTitleSavesTitleInSettings(self):
        self.viewer.setTitle('New title')
        self.assertEqual('New title', self.settings.get(self.viewer.settingsSection(), 'title'))
        
    def testSetTitleDoesNotSaveTitleInSettingsWhenTitleIsDefaultTitle(self):
        self.viewer.setTitle(self.viewer.defaultTitle)
        self.assertEqual('', self.settings.get(self.viewer.settingsSection(), 'title'))

    def testSetTitleChangesTabTitle(self):
        self.viewer.setTitle('New title')
        self.assertEqual('New title', self.window.manager.GetPane(self.viewer).caption)

    def testGetItemTooltipData(self):
        self.task.setDescription('Description')
        expectedData = [('led_grey_icon', ['task']), ('folder_blue_arrow_icon', []), ('paperclip_icon', []), (None, ['Description'])]
        self.assertEqual(expectedData, self.viewer.getItemTooltipData(self.task))


class SortableViewerTest(test.TestCase):
    def setUp(self):
        self.settings = config.Settings(load=False)
        self.viewer = self.createViewer()
        
    def createViewer(self):
        viewer = gui.viewer.mixin.SortableViewerMixin()
        viewer.settings = self.settings
        viewer.settingsSection = lambda: 'taskviewer'
        viewer.SorterClass = task.sorter.Sorter
        presentation = viewer.createSorter(task.TaskList())
        viewer.presentation = lambda: presentation
        return viewer
        
    def testIsSortable(self):
        self.failUnless(self.viewer.isSortable())

    def testSortBy(self):
        self.viewer.sortBy('subject')
        self.assertEqual('subject',
            eval(self.settings.get(self.viewer.settingsSection(), 'sortby'))[0])

    def testSortByTwiceFlipsSortOrder(self):
        self.viewer.sortBy('subject')
        self.viewer.setSortOrderAscending(True)
        self.viewer.sortBy('subject')
        self.failIf(self.viewer.isSortOrderAscending())

    def testIsSortedBy(self):
        self.viewer.sortBy('description')
        self.failUnless(self.viewer.isSortedBy('description'))
        
    def testSortOrderAscending(self):
        self.viewer.setSortOrderAscending(True)
        self.failUnless(self.viewer.isSortOrderAscending())

    def testSortOrderDescending(self):
        self.viewer.setSortOrderAscending(False)
        self.failIf(self.viewer.isSortOrderAscending())
                
    def testSetSortCaseSensitive(self):
        self.viewer.setSortCaseSensitive(True)
        self.failUnless(self.viewer.isSortCaseSensitive())
        
    def testSetSortCaseInsensitive(self):
        self.viewer.setSortCaseSensitive(False)
        self.failIf(self.viewer.isSortCaseSensitive())
        
    def testApplySettingsWhenCreatingViewer(self):
        self.settings.set(self.viewer.settingsSection(), 'sortby', '["description"]')
        anotherViewer = self.createViewer()
        anotherViewer.presentation().extend([task.Task(description='B'), 
                                             task.Task(description='A')])
        self.assertEqual(['A', 'B'], 
                         [t.description() for t in anotherViewer.presentation()])


class SortableViewerForTasksTest(test.TestCase):
    def setUp(self):
        self.settings = config.Settings(load=False)
        class ViewerUnderTest(gui.viewer.mixin.SortableViewerForTasksMixin):
            viewerImages = []
        self.viewer = ViewerUnderTest()
        self.viewer.settings = self.settings
        self.viewer.settingsSection = lambda: 'taskviewer'
        self.viewer.presentation = lambda: task.sorter.Sorter(task.TaskList())

    def testSetSortByTaskStatusFirst(self):
        self.viewer.setSortByTaskStatusFirst(True)
        self.failUnless(self.viewer.isSortByTaskStatusFirst())
        
    def testSetNoSortByTaskStatusFirst(self):
        self.viewer.setSortByTaskStatusFirst(False)
        self.failIf(self.viewer.isSortByTaskStatusFirst())

      
class DummyViewer(object):
    def isTreeViewer(self):
        return False
    
    def createFilter(self, presentation):
        return presentation


class SearchableViewerUnderTest(gui.viewer.mixin.SearchableViewerMixin, 
                                DummyViewer):
    pass   

    
class SearchableViewerTest(test.TestCase):
    def setUp(self):
        self.settings = config.Settings(load=False)
        self.viewer = self.createViewer()
        
    def createViewer(self):
        viewer = SearchableViewerUnderTest()
        # pylint: disable=W0201
        viewer.settings = self.settings
        viewer.settingsSection = lambda: 'taskviewer'
        presentation = viewer.createFilter(task.TaskList())
        viewer.presentation = lambda: presentation
        return viewer
        
    def testIsSearchable(self):
        self.failUnless(self.viewer.isSearchable())
        
    def testDefaultSearchFilter(self):
        self.assertEqual(('', False, False, False, False), self.viewer.getSearchFilter())
        
    def testSetSearchFilterString(self):
        self.viewer.setSearchFilter('bla', matchCase=True)
        self.assertEqual('bla', self.settings.get(self.viewer.settingsSection(),
                                                  'searchfilterstring'))

    def testSetSearchFilterString_AffectsPresentation(self):
        self.viewer.presentation().append(task.Task())
        self.viewer.setSearchFilter('bla')
        self.failIf(self.viewer.presentation())
        
    def testSearchMatchCase(self):
        self.viewer.setSearchFilter('bla', matchCase=True)
        self.assertEqual(True, 
            self.settings.getboolean(self.viewer.settingsSection(), 
                                     'searchfiltermatchcase'))
        
    def testSearchMatchCase_AffectsPresenation(self):
        self.viewer.presentation().append(task.Task('BLA'))
        self.viewer.setSearchFilter('bla', matchCase=True)
        self.failIf(self.viewer.presentation())
        
    def testSearchIncludesSubItems(self):
        self.viewer.setSearchFilter('bla', includeSubItems=True)
        self.assertEqual(True, 
            self.settings.getboolean(self.viewer.settingsSection(), 
                                     'searchfilterincludesubitems'))
        
    def testSearchIncludesSubItems_AffectsPresentation(self):
        parent = task.Task('parent')
        child = task.Task('child')
        parent.addChild(child)
        self.viewer.presentation().append(parent)
        self.viewer.setSearchFilter('parent', includeSubItems=True)
        self.assertEqual(2, len(self.viewer.presentation()))
        
    def testSearchDescription(self):
        self.viewer.setSearchFilter('bla', searchDescription=True)
        self.assertEqual(True, self.settings.getboolean(self.viewer.settingsSection(),
                                                        'searchdescription'))
        
    def testSearchDescription_AffectsPresentation(self):
        self.viewer.presentation().append(task.Task('subject', description='description'))
        self.viewer.setSearchFilter('descr', searchDescription=True)
        self.assertEqual(1, len(self.viewer.presentation()))
        
    def testApplySettingsWhenCreatingViewer(self):
        self.settings.set(self.viewer.settingsSection(), 'searchfilterstring', 'whatever')
        anotherViewer = self.createViewer()
        anotherViewer.presentation().append(task.Task())
        self.failIf(anotherViewer.presentation())


class FilterableViewerTest(test.TestCase):
    def setUp(self):
        self.viewer = gui.viewer.mixin.FilterableViewerMixin()
        
    def testIsFilterable(self):
        self.failUnless(self.viewer.isFilterable())


class FilterableViewerForTasksUnderTest(gui.viewer.mixin.FilterableViewerForTasksMixin, 
                                        DummyViewer):
    pass
        
        
class FilterableViewerForTasks(test.TestCase):
    def setUp(self):
        self.settings = config.Settings(load=False)
        task.Task.settings = self.settings
        self.viewer = self.createViewer()

    def tearDown(self):
        super(FilterableViewerForTasks, self).tearDown()
        self.viewer.taskFile.close()
        self.viewer.taskFile.stop()

    def createViewer(self):
        viewer = FilterableViewerForTasksUnderTest()
        # pylint: disable=W0201
        viewer.taskFile = persistence.TaskFile()
        viewer.settings = self.settings
        viewer.settingsSection = lambda: 'taskviewer'
        presentation = viewer.createFilter(viewer.taskFile.tasks())
        viewer.presentation = lambda: presentation
        return viewer

    def testIsNotHidingInactiveTasksByDefault(self):
        self.failIf(self.viewer.isHidingTaskStatus(task.status.inactive))

    def testHideInactiveTasks(self):
        self.viewer.hideTaskStatus(task.status.inactive)
        self.failUnless(self.viewer.isHidingTaskStatus(task.status.inactive))
        
    def testHideInactiveTasks_SetsSetting(self):
        self.viewer.hideTaskStatus(task.status.inactive)
        self.failUnless(self.settings.getboolean(self.viewer.settingsSection(), 
                                                 'hideinactivetasks'))

    def testHideInactiveTasks_AffectsPresentation(self):
        self.viewer.presentation().append(task.Task(plannedStartDateTime=date.Tomorrow()))
        self.viewer.hideTaskStatus(task.status.inactive)
        self.failIf(self.viewer.presentation())
    
    def testUnhideInactiveTasks(self):
        self.viewer.presentation().append(task.Task(plannedStartDateTime=date.Tomorrow()))
        self.viewer.hideTaskStatus(task.status.inactive)
        self.viewer.hideTaskStatus(task.status.inactive, False)
        self.failUnless(self.viewer.presentation())

    def testIsNotHidingLateTasksByDefault(self):
        self.failIf(self.viewer.isHidingTaskStatus(task.status.late))

    def testHideLateTasks(self):
        self.viewer.hideTaskStatus(task.status.late)
        self.failUnless(self.viewer.isHidingTaskStatus(task.status.late))
        
    def testHideLateTasks_SetsSetting(self):
        self.viewer.hideTaskStatus(task.status.late)
        self.failUnless(self.settings.getboolean(self.viewer.settingsSection(), 
                                                 'hidelatetasks'))

    def testHideLateTasks_AffectsPresentation(self):
        self.viewer.presentation().append(task.Task(plannedStartDateTime=date.Yesterday()))
        self.viewer.hideTaskStatus(task.status.late)
        self.failIf(self.viewer.presentation())
    
    def testUnhideLateTasks(self):
        self.viewer.presentation().append(task.Task(plannedStartDateTime=date.Yesterday()))
        self.viewer.hideTaskStatus(task.status.late)
        self.viewer.hideTaskStatus(task.status.late, False)
        self.failUnless(self.viewer.presentation())
 
    def testIsNotHidingDueSoonTasksByDefault(self):
        self.failIf(self.viewer.isHidingTaskStatus(task.status.duesoon))

    def testHideDueSoonTasks(self):
        self.viewer.hideTaskStatus(task.status.duesoon)
        self.failUnless(self.viewer.isHidingTaskStatus(task.status.duesoon))
        
    def testHideDueSoonTasks_SetsSetting(self):
        self.viewer.hideTaskStatus(task.status.duesoon)
        self.failUnless(self.settings.getboolean(self.viewer.settingsSection(), 
                                                 'hideduesoontasks'))

    def testHideDueSoonTasks_AffectsPresentation(self):
        self.viewer.presentation().append(task.Task(dueDateTime=date.Now() + date.ONE_HOUR))
        self.viewer.hideTaskStatus(task.status.duesoon)
        self.failIf(self.viewer.presentation())
    
    def testUnhideDueSoonTasks(self):
        self.viewer.presentation().append(task.Task(dueDateTime=date.Now() + date.ONE_HOUR))
        self.viewer.hideTaskStatus(task.status.duesoon)
        self.viewer.hideTaskStatus(task.status.duesoon, False)
        self.failUnless(self.viewer.presentation())

    def testIsNotHidingOverDueTasksByDefault(self):
        self.failIf(self.viewer.isHidingTaskStatus(task.status.overdue))

    def testHideOverDueTasks(self):
        self.viewer.hideTaskStatus(task.status.overdue)
        self.failUnless(self.viewer.isHidingTaskStatus(task.status.overdue))
        
    def testHideOverDueTasks_SetsSetting(self):
        self.viewer.hideTaskStatus(task.status.overdue)
        self.failUnless(self.settings.getboolean(self.viewer.settingsSection(), 
                                                 'hideoverduetasks'))

    def testHideOverDueTasks_AffectsPresentation(self):
        self.viewer.presentation().append(task.Task(dueDateTime=date.Yesterday()))
        self.viewer.hideTaskStatus(task.status.overdue)
        self.failIf(self.viewer.presentation())
    
    def testUnhideOverDueTasks(self):
        self.viewer.presentation().append(task.Task(dueDateTime=date.Yesterday()))
        self.viewer.hideTaskStatus(task.status.overdue)
        self.viewer.hideTaskStatus(task.status.overdue, False)
        self.failUnless(self.viewer.presentation())
       
    def testIsNotHidingCompletedTasksByDefault(self):
        self.failIf(self.viewer.isHidingTaskStatus(task.status.completed))
        
    def testHideCompletedTasks(self):
        self.viewer.hideTaskStatus(task.status.completed)
        self.failUnless(self.viewer.isHidingTaskStatus(task.status.completed))
    
    def testHideCompletedTasks_SetsSetting(self):
        self.viewer.hideTaskStatus(task.status.completed)
        self.failUnless(self.settings.getboolean(self.viewer.settingsSection(),
                                                 'hidecompletedtasks'))
    
    def testHideCompletedTasks_AffectsPresentation(self):
        self.viewer.presentation().append(task.Task(completionDateTime=date.Now()))
        self.viewer.hideTaskStatus(task.status.completed)
        self.failIf(self.viewer.presentation())
        
    def testUnhideCompletedTasks(self):    
        self.viewer.presentation().append(task.Task(completionDateTime=date.Now()))
        self.viewer.hideTaskStatus(task.status.completed)
        self.viewer.hideTaskStatus(task.status.completed, False)
        self.failUnless(self.viewer.presentation())

    def testIsNotHidingCompositeTasksByDefault(self):
        self.failIf(self.viewer.isHidingCompositeTasks())
        
    def testHideCompositeTasks(self):
        self.viewer.hideCompositeTasks()
        self.failUnless(self.viewer.isHidingCompositeTasks())
        
    def testHideCompositeTasks_SetsSettings(self):
        self.viewer.hideCompositeTasks()
        self.failUnless(self.settings.getboolean(self.viewer.settingsSection(),
                                                 'hidecompositetasks'))

    def testHideCompositeTasks_AffectsPresentation(self):
        self.viewer.hideCompositeTasks()
        parent = task.Task()
        child = task.Task()
        parent.addChild(child)
        self.viewer.presentation().append(parent)
        self.assertEqual([child], self.viewer.presentation())
        
    def testUnhideCompositeTasks(self):
        self.viewer.hideCompositeTasks()
        parent = task.Task()
        child = task.Task()
        parent.addChild(child)
        self.viewer.presentation().append(parent)
        self.viewer.hideCompositeTasks(False)
        self.assertEqual(2, len(self.viewer.presentation()))
        
    def testClearAllFilters(self):
        self.viewer.hideCompositeTasks()
        for status in task.Task.possibleStatuses():
            self.viewer.hideTaskStatus(status)
        self.viewer.resetFilter()
        self.failIf(self.viewer.isHidingCompositeTasks())
        for status in task.Task.possibleStatuses():
            self.failIf(self.viewer.isHidingTaskStatus(status))
        
    def testApplySettingsWhenCreatingViewer(self):
        self.settings.set(self.viewer.settingsSection(), 'hidecompletedtasks', 'True')
        anotherViewer = self.createViewer()
        anotherViewer.presentation().append(task.Task(completionDateTime=date.Now()))
        self.failIf(anotherViewer.presentation())   


class ViewerBaseClassTest(test.wxTestCase):
    def testNotImplementedError(self):
        taskFile = persistence.TaskFile()
        try:
            try:
                gui.viewer.base.Viewer(self.frame, taskFile,
                                       None, settingsSection='bla')
                self.fail('Expected NotImplementedError') # pragma: no cover
            except NotImplementedError:
                pass
        finally:
            taskFile.close()
            taskFile.stop()


class ViewerIteratorTestCase(test.wxTestCase):
    treeMode = 'Subclass responsibility'
    
    def createViewer(self):
        return gui.viewer.TaskViewer(self.window, self.taskFile,
            self.settings)

    def setUp(self):
        super(ViewerIteratorTestCase, self).setUp()
        self.settings = config.Settings(load=False)
        task.Task.settings = self.settings
        self.taskFile = persistence.TaskFile()
        self.taskList = self.taskFile.tasks()
        self.window = AuiManagedFrameWithDynamicCenterPane(self.frame)
        self.viewer = self.createViewer()
        self.settings.setboolean(self.viewer.settingsSection(), 'treemode',
                                 self.treeMode == 'True')
        self.viewer.sortBy('subject')

    def tearDown(self):
        super(ViewerIteratorTestCase, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()

    def getItemsFromIterator(self):
        return list(self.viewer.visibleItems()) # pylint: disable=E1101


class ViewerIteratorTestsMixin(object):
    def testEmptyPresentation(self):
        self.assertEqual([], self.getItemsFromIterator())
        
    def testOneItem(self):
        self.taskList.append(task.Task())
        self.assertEqual(self.taskList, self.getItemsFromIterator())
        
    def testOneParentAndOneChild(self):
        parent = task.Task('Z')
        child = task.Task('A', parent=parent)
        parent.addChild(child)
        self.taskList.append(parent)
        if self.treeMode == 'True':
            expectedParentAndChildOrder = [parent, child]
        else:
            expectedParentAndChildOrder = [child, parent]
        self.assertEqual(expectedParentAndChildOrder, 
                         self.getItemsFromIterator())

    def testOneParentOneChildAndOneGrandChild(self):
        parent = task.Task('a-parent')
        child = task.Task('b-child', parent=parent)
        grandChild = task.Task('c-grandchild', parent=child)
        parent.addChild(child)
        child.addChild(grandChild)
        self.taskList.append(parent)
        self.assertEqual([parent, child, grandChild], 
                         self.getItemsFromIterator())
    
    def testThatTasksNotInPresentationAreExcluded(self):
        parent = task.Task('parent')
        child = task.Task('child')
        parent.addChild(child)
        self.taskList.append(parent)
        self.viewer.setSearchFilter('parent', matchCase=True)
        self.assertEqual([parent], self.getItemsFromIterator())
        
    
class TreeViewerIteratorTest(ViewerIteratorTestCase, ViewerIteratorTestsMixin):
    treeMode = 'True'
        
        
class ListViewerIteratorTest(ViewerIteratorTestCase, ViewerIteratorTestsMixin):
    treeMode = 'False'



class ViewerWithColumnsTest(test.wxTestCase):
    def setUp(self):
        self.settings = config.Settings(load=False)
        self.taskFile = persistence.TaskFile()
        self.viewer = gui.viewer.TaskViewer(self.frame, self.taskFile, self.settings)

    def tearDown(self):
        super(ViewerWithColumnsTest, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()

    def testDefaultColumnWidth(self):
        expectedWidth = hypertreelist._DEFAULT_COL_WIDTH # pylint: disable=W0212
        self.assertEqual(expectedWidth, self.viewer.getColumnWidth('subject'))
        



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from __future__ import with_statement
import test, tempfile
from taskcoachlib.thirdparty import lockfile


class LockFileTest(test.TestCase):
    def setUp(self):
        self.tmpfile = tempfile.NamedTemporaryFile()
        self.lock = lockfile.FileLock(self.tmpfile.name)
        
    def tearDown(self):
        super(LockFileTest, self).tearDown()
        self.tmpfile.close() # Temp files are deleted when closed
        
    def testFileIsNotLockedInitially(self):
        self.failIf(self.lock.is_locked())
        
    def testFileIsLockedAfterLocking(self):
        self.lock.acquire()
        self.failUnless(self.lock.is_locked())
        
    def testLockingWithContextManager(self):
        with self.lock:
            self.failUnless(self.lock.is_locked())
        self.failIf(self.lock.is_locked())

    def testLockingTwoFiles(self):
        self.lock.acquire()
        tmpfile2 = tempfile.NamedTemporaryFile()
        lock2 = lockfile.FileLock(tmpfile2.name)
        lock2.acquire()
        self.failUnless(self.lock.is_locked())
        self.failUnless(lock2.is_locked())



# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

# pylint: disable=C0103

# Edit these for every release:

version = '1.4.3'  # Current version number of the application
tskversion = 37  # Current version number of the task file format, changed to 37 for release 1.3.23.
release_day = '31'  # Day number of the release, 1-31, as string
release_month = 'January'  # Month of the release in plain English
release_year = '2016'  # Year of the release as string
release_status = 'stable'  # One of 'alpha', 'beta', 'stable'

# No editing needed below this line for doing a release.

import re, datetime
try:
    from taskcoachlib.meta.revision import revision  # pylint: disable=F0401,W0611
except ImportError:
    revision = None 

months = ['January', 'February', 'March', 'April', 'May', 'June', 
          'July', 'August', 'September', 'October', 'November', 'December']

if revision:  # Buildbot sets revision
    # Decrement version because this version isn't released yet. This
    # assumes that version components are < 100; 99 will actually mean
    # pre-major release
    # pylint: disable=W0141
    major, inter, minor = map(int, version.split('.'))
    numversion = major * 10000 + inter * 100 + minor
    numversion -= 1
    major = numversion // 10000
    inter = (numversion // 100) % 100
    minor = numversion % 100
    version = '.'.join(map(str, [major, inter, minor]))

    now = datetime.datetime.today()
    release_day = str(now.day)
    release_month = months[now.month-1]
    release_year = str(now.year)
    release_status = 'beta'
    version += '.' + revision

assert release_month in months  # Try to prevent typo's
release_month_nr = '%02d' % (months.index(release_month) + 1)
release_day_nr = '%02d' % int(release_day)
date = release_month + ' ' + release_day + ', ' + release_year

name = 'Task Coach'
description = 'Your friendly task manager'
long_description = '%(name)s is a free open source todo manager. It grew ' \
'out of frustration about other programs not handling composite tasks well. ' \
'In addition to flexible composite tasks, %(name)s has grown to include ' \
'prerequisites, prioritizing, effort tracking, category tags, budgets, ' \
'notes, and many other features. However, users are not forced to use all ' \
'these features; %(name)s can be as simple or complex as you need it to be. ' \
'%(name)s is available for Windows, Mac OS X, and GNU/Linux; and there is a ' \
'companion iOS app.' % dict(name=name)
keywords = 'task manager, todo list, pim, time registration, track effort'
author_first, author_last = 'Frank', 'Niessink'  # Needed for PAD file
author = '%s %s, Jerome Laheurte, and Aaron Wolf' % (author_first, author_last)
author_unicode = u'%s %s, Jérôme Laheurte, and Aaron Wolf' % (author_first, author_last)
author_email = 'developers@taskcoach.org'

filename = name.replace(' ', '')
filename_lower = filename.lower()

url = 'http://taskcoach.org/'  # Don't remove the trailing slash, other code is assuming it will be there
screenshot = url + 'screenshots/Windows/0.71.2-Windows_XP-Tasks_categories_and_effort.png'
icon = url + 'taskcoach.png'
pad = url + 'pad.xml'
version_url = url + 'version.txt'
message_url = url + 'messages.txt'
download = url + 'download.html'
dist_download_prefix = 'http://downloads.sourceforge.net/%s' % filename_lower
faq_url = 'https://answers.launchpad.net/taskcoach/+faqs'
sf_tracker_url = 'https://sourceforge.net/tracker/'
bug_report_url = sf_tracker_url + '?func=add&group_id=130831&atid=719134'
known_bugs_url = sf_tracker_url + '?group_id=130831&atid=719134&status=1'
support_request_url = sf_tracker_url + '?group_id=130831&atid=719135'
feature_request_url = 'http://taskcoach.uservoice.com'
translations_url = 'https://translations.launchpad.net/taskcoach'
donate_url = url + 'givesupport.html' 
i18n_url = url + 'i18n.html'

announcement_addresses = 'taskcoach@yahoogroups.com, python-announce-list@python.org'
bcc_announcement_addresses = 'johnhaller@portableapps.com'

copyright = 'Copyright (C) 2004-%s %s'%(release_year, author)  # pylint: disable=W0622
license_title = 'GNU General Public License'
license_version = '3'
license_title_and_version = '%s version %s'%(license_title, license_version) 
license = '%s or any later version'%license_title_and_version  # pylint: disable=W0622
license_title_and_version_abbrev = 'GPLv%s'%license_version
license_abbrev = '%s+'%license_title_and_version_abbrev
license_notice = '''%(name)s is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

%(name)s is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.''' % dict(name=name)

license_notice_html = '<p>%s</p>'%license_notice.replace('\n\n', '</p><p>')
license_notice_html = re.sub(r'<http([^>]*)>', r'<a href="http\1" target="_blank">http\1</a>', license_notice_html)

platform = 'Any'
pythonversion = '2.6'
wxpythonversionnumber = '2.8.9.2'
wxpythonversion = '%s-unicode' % wxpythonversionnumber
twistedversionnumber = '10.0'

languages = {
    'English (US)': (None, True),
    'English (AU)': ('en_AU', True),
    'English (CA)': ('en_CA', True),
    'English (GB)': ('en_GB', True),
    'Arabic': ('ar', False),
    'Basque': ('eu', False),
    'Belarusian': ('be', False),
    'Bosnian': ('bs', False),
    'Breton': ('br', False),
    'Bulgarian': ('bg', False),
    'Catalan': ('ca', False),
    'Chinese (Simplified)': ('zh_CN', False),
    'Chinese (Traditional)': ('zh_TW', False),
    'Czech': ('cs', True),
    'Danish': ('da', False),
    'Dutch': ('nl', True),
    'Esperanto': ('eo', False),
    'Estonian': ('et', False),
    'Finnish': ('fi', True),
    'French': ('fr', True), 
    'Galician': ('gl', False),
    'German': ('de', True),
    'German (Low)': ('nds', False),
    'Greek': ('el', False),
    'Hebrew': ('he', False),
    'Hindi': ('hi', False),
    'Hungarian': ('hu', False),
    'Indonesian': ('id', False),
    'Italian': ('it', True),
    'Japanese': ('ja', False),
    'Korean': ('ko', False),
    'Latvian': ('lv', False),
    'Lithuanian': ('lt', False),
    'Marathi': ('mr', False),
    'Mongolian': ('mn', False),
    'Norwegian (Bokmal)': ('nb', False),
    'Norwegian (Nynorsk)': ('nn', False),
    'Occitan': ('oc', False),
    'Papiamento': ('pap', False),
    'Persian': ('fa', False),
    'Polish': ('pl', True),
    'Portuguese': ('pt', True),
    'Portuguese (Brazilian)': ('pt_BR', True),
    'Romanian': ('ro', True),
    'Russian': ('ru', True),
    'Slovak': ('sk', True),
    'Slovene': ('sl', False),
    'Spanish': ('es', True),
    'Swedish': ('sv', False),
    'Telugu': ('te', False),
    'Thai': ('th', False),
    'Turkish': ('tr', True),
    'Ukranian': ('uk', False),
    'Vietnamese': ('vi', False)}
languages_list = ','.join(languages.keys())

def __createDict(localsDict):
    ''' Provide the local variables as a dictionary for use in string
        formatting. '''
    metaDict = {}  # pylint: disable=W0621
    for key in localsDict:
        if not key.startswith('__'):
            metaDict[key] = localsDict[key]
    return metaDict

metaDict = __createDict(locals())



# -*- coding: UTF-8 -*-
'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib.gui.dialog import entry 


class AmountEntryTest(test.wxTestCase):
    def setUp(self):
        super(AmountEntryTest, self).setUp()
        self.amountEntry = entry.AmountEntry(self.frame)

    def testCreate(self):
        self.assertEqual(0.0, self.amountEntry.GetValue())

    def testSetValue(self):
        self.amountEntry.SetValue(1.0)
        self.assertEqual(1.0, self.amountEntry.GetValue())



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
from taskcoachlib.domain.task import Task
from taskcoachlib import persistence, operating_system
from taskcoachlib.i18n import _
from taskcoachlib.thirdparty.deltaTime import nlTimeExpression
from wx.lib import sized_controls


class TimeExpressionEntry(wx.TextCtrl):
    def __init__(self, *args, **kwargs):
        super(TimeExpressionEntry, self).__init__(*args, **kwargs)

        self.__defaultColor = self.GetBackgroundColour()
        self.__invalidColor = wx.Colour(255, 128, 128)

        wx.EVT_TEXT(self, wx.ID_ANY, self._onTextChanged)

    @staticmethod
    def isValid(value):
        if value:
            try:
                res = nlTimeExpression.parseString(value)
            except:
                return False # pylint: disable=W0702
            return 'calculatedTime' in res
        return True # Empty is valid.

    def _onTextChanged(self, event):
        event.Skip()
        self.SetBackgroundColour(self.__defaultColor if self.isValid(self.GetValue()) else self.__invalidColor)


class TemplatesDialog(sized_controls.SizedDialog):
    def __init__(self, settings, *args, **kwargs):
        self.settings = settings
        self._changing = False
        super(TemplatesDialog, self).__init__(style=wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER, 
                                              *args, **kwargs)
        pane = self.GetContentsPane()
        pane.SetSizerType('vertical')
        self.createInterior(pane)
        self._buttonSizer = self.CreateStdDialogButtonSizer(wx.OK|wx.CANCEL)
        self.SetButtonSizer(self._buttonSizer)
        self.Fit()
        self.SetMinSize(self.GetSize()) # Current size is min size
        self._buttonSizer.GetAffirmativeButton().Bind(wx.EVT_BUTTON, self.ok)
        self.CentreOnParent()
    
    def createInterior(self, pane):
        self.createTemplateList(pane)
        self.createTemplateEntries(pane)
        
    def createTemplateList(self, pane):
        panel = sized_controls.SizedPanel(pane)
        panel.SetSizerType('horizontal')
        panel.SetSizerProps(expand=True, proportion=1)
        self._templateList = wx.TreeCtrl(panel, style=wx.TR_HAS_BUTTONS|wx.TR_HIDE_ROOT|wx.TR_SINGLE)
        self._templateList.SetMinSize((300, 200))
        self._templateList.SetSizerProps(expand=True, proportion=1)
        self._templateList.Bind(wx.EVT_TREE_SEL_CHANGED, self.OnSelectionChanged)
        self._templates = persistence.TemplateList(self.settings.pathToTemplatesDir())
        self._root = self._templateList.AddRoot('Root')
        for task in self._templates.tasks():
            item = self.appendTemplate(self._root, task)
            if operating_system.isMac():
                # See http://trac.wxwidgets.org/ticket/10085
                self._templateList.SetItemText(item, task.subject())
        self.createTemplateListButtons(panel)
        panel.Fit()
        
    def createTemplateListButtons(self, pane):
        panel = sized_controls.SizedPanel(pane)
        panel.SetSizerType('vertical')
        self._btnDelete = self.createButton(panel, 'cross_red_icon', self.OnDelete, enable=False)
        self._btnUp = self.createButton(panel, 'arrow_up_icon', self.OnUp, enable=False) 
        self._btnDown = self.createButton(panel, 'arrow_down_icon', self.OnDown, enable=False)
        self._btnAdd = self.createButton(panel, 'symbol_plus_icon', self.OnAdd)
        panel.Fit()
        
    def createButton(self, parent, bitmapName, handler, enable=True):
        bitmap = wx.ArtProvider_GetBitmap(bitmapName, size=(32, 32))
        button = wx.BitmapButton(parent, bitmap=bitmap)
        button.Bind(wx.EVT_BUTTON, handler)
        button.Enable(enable)
        return button

    def createTemplateEntries(self, pane):
        panel = self._editPanel = sized_controls.SizedPanel(pane)
        panel.SetSizerType('form')
        panel.SetSizerProps(expand=True)
        label = wx.StaticText(panel, label=_('Subject'))
        label.SetSizerProps(valign='center')
        self._subjectCtrl = wx.TextCtrl(panel)
        label = wx.StaticText(panel, label=_('Planned start date'))
        label.SetSizerProps(valign='center')
        self._plannedStartDateTimeCtrl = TimeExpressionEntry(panel)
        label = wx.StaticText(panel, label=_('Due date'))
        label.SetSizerProps(valign='center')
        self._dueDateTimeCtrl = TimeExpressionEntry(panel)
        label = wx.StaticText(panel, label=_('Completion date'))
        label.SetSizerProps(valign='center')
        self._completionDateTimeCtrl = TimeExpressionEntry(panel)
        label = wx.StaticText(panel, label=_('Reminder'))
        label.SetSizerProps(valign='center')
        self._reminderDateTimeCtrl = TimeExpressionEntry(panel)
        self._taskControls = (self._subjectCtrl, self._plannedStartDateTimeCtrl, self._dueDateTimeCtrl,
                              self._completionDateTimeCtrl, self._reminderDateTimeCtrl)
        for ctrl in self._taskControls:
            ctrl.SetSizerProps(valign='center', expand=True)
            ctrl.Bind(wx.EVT_TEXT, self.onValueChanged)
        self.enableEditPanel(False)
        panel.Fit()
            
    def enableEditPanel(self, enable=True):
        for ctrl in self._taskControls:
            ctrl.Enable(enable)
            
    def appendTemplate(self, parentItem, task):
        item = self._templateList.AppendItem(parentItem, task.subject(), data=wx.TreeItemData(task))
        for child in task.children():
            self.appendTemplate(item, child)
        return item

    def onValueChanged(self, event):
        event.Skip()
        if self._GetSelection().IsOk() and not self._changing:
            task = self._templateList.GetItemData(self._GetSelection()).GetData()
            task.setSubject(self._subjectCtrl.GetValue())
            for ctrl, name in [(self._plannedStartDateTimeCtrl, 'plannedstartdatetmpl'),
                               (self._dueDateTimeCtrl, 'duedatetmpl'),
                               (self._completionDateTimeCtrl, 'completiondatetmpl'),
                               (self._reminderDateTimeCtrl, 'remindertmpl')]:
                if TimeExpressionEntry.isValid(ctrl.GetValue()):
                    setattr(task, name, ctrl.GetValue() or None)

    def _GetSelection(self):
        return self._templateList.GetSelection()

    def OnSelectionChanged(self, event): # pylint: disable=W0613
        self._changing = True
        try:
            selection = self._GetSelection()
            selectionOK = selection.IsOk() and selection != self._root
            selectionAtRoot = False
            if selectionOK:
                selectionAtRoot = (self._templateList.GetItemParent(selection) == self._root)
            self._btnDelete.Enable(selectionAtRoot)
            self._btnUp.Enable(selectionAtRoot and self._templateList.GetPrevSibling(selection).IsOk())
            self._btnDown.Enable(selectionAtRoot and self._templateList.GetNextSibling(selection).IsOk())
            self.enableEditPanel(selectionOK)
            if selectionOK:
                task = self._templateList.GetItemData(selection).GetData()
                if task is None:
                    for ctrl in self._taskControls:
                        ctrl.SetValue(u'')
                else:
                    self._subjectCtrl.SetValue(task.subject())
                    self._plannedStartDateTimeCtrl.SetValue(task.plannedstartdatetmpl or u'')
                    self._dueDateTimeCtrl.SetValue(task.duedatetmpl or u'')
                    self._completionDateTimeCtrl.SetValue(task.completiondatetmpl or u'')
                    self._reminderDateTimeCtrl.SetValue(task.remindertmpl or u'')
            else:
                for ctrl in self._taskControls:
                    ctrl.SetValue(u'')
        finally:
            self._changing = False

    def OnDelete(self, event): # pylint: disable=W0613
        task = self._templateList.GetItemData(self._GetSelection()).GetData()
        index = self._templates.tasks().index(task)
        self._templates.deleteTemplate(index)
        self._templateList.Delete(self._GetSelection())

    def OnUp(self, event): # pylint: disable=W0613
        selection = self._GetSelection()
        prev = self._templateList.GetPrevSibling(selection)
        prev = self._templateList.GetPrevSibling(prev)
        task = self._templateList.GetItemData(selection).GetData()
        self._templateList.Delete(selection)
        if prev.IsOk():
            item = self._templateList.InsertItem(self._root, prev, task.subject(), data=wx.TreeItemData(task))
        else:
            item = self._templateList.PrependItem(self._root, task.subject(), data=wx.TreeItemData(task))
        for child in task.children():
            self.appendTemplate(item, child)
        index = self._templates.tasks().index(task)
        self._templates.swapTemplates(index - 1, index)
        self._templateList.SelectItem(item)

    def OnDown(self, event): # pylint: disable=W0613
        selection = self._GetSelection()
        next = self._templateList.GetNextSibling(selection)
        task = self._templateList.GetItemData(selection).GetData()
        self._templateList.Delete(selection)
        item = self._templateList.InsertItem(self._root, next, task.subject(), data=wx.TreeItemData(task))
        for child in task.children():
            self.appendTemplate(item, child)
        index = self._templates.tasks().index(task)
        self._templates.swapTemplates(index, index + 1)
        self._templateList.SelectItem(item)

    def OnAdd(self, event): # pylint: disable=W0613
        template = Task(subject=_('New task template'))
        for name in ('plannedstartdatetmpl', 'duedatetmpl', 'completiondatetmpl', 
                     'remindertmpl'):
            setattr(template, name, None)
        theTask = self._templates.addTemplate(template)
        self.appendTemplate(self._root, theTask)

    def ok(self, event):
        self._templates.save()
        event.Skip()
            



import codecs
import base64

from ..py30compat import unittest
from ..test_backend import BackendBasicTests
from keyring.backends import Google
from keyring.credentials import SimpleCredential
from keyring.backend import NullCrypter
from keyring import errors
from keyring.py27compat import input, pickle
from .. import mocks

def is_gdata_supported():
    try:
        __import__('gdata.service')
    except ImportError:
        return False
    return True

def init_google_docs_keyring(client, can_create=True,
                             input_getter=input):
    credentials = SimpleCredential('foo', 'bar')
    return Google.DocsKeyring(credentials,
                                             'test_src',
                                             NullCrypter(),
                                             client=client,
                                             can_create=can_create,
                                             input_getter=input_getter
                                            )

@unittest.skipUnless(is_gdata_supported(),
                     "Need Google Docs (gdata)")
class GoogleDocsKeyringTestCase(BackendBasicTests, unittest.TestCase):
    """Run all the standard tests on a new keyring"""

    def init_keyring(self):
        client = mocks.MockDocumentService()
        client.SetClientLoginToken('foo')
        return init_google_docs_keyring(client)

@unittest.skipUnless(is_gdata_supported(),
                     "Need Google Docs (gdata)")
class GoogleDocsKeyringInteractionTestCase(unittest.TestCase):
    """Additional tests for Google Doc interactions"""

    def _init_client(self, set_token=True):
        client = mocks.MockDocumentService()
        if set_token:
            client.SetClientLoginToken('interaction')
        return client

    def _init_keyring(self, client):
        self.keyring = init_google_docs_keyring(client)

    def _init_listfeed(self):
        listfeed = mocks.MockListFeed()
        listfeed._entry = [mocks.MockDocumentListEntry(),
                           mocks.MockDocumentListEntry()
                          ]
        return listfeed

    def _encode_data(self, data):
        return base64.urlsafe_b64encode(pickle.dumps(data))

    def test_handles_auth_failure(self):
        import gdata
        client = self._init_client(set_token=False)
        client._login_err = gdata.service.BadAuthentication
        self._init_keyring(client)
        with self.assertRaises(errors.InitError):
            self.keyring.client

    def test_handles_auth_error(self):
        import gdata
        client = self._init_client(set_token=False)
        client._login_err = gdata.service.Error
        self._init_keyring(client)
        with self.assertRaises(errors.InitError):
            self.keyring.client

    def test_handles_login_captcha(self):
        import gdata
        client = self._init_client(set_token=False)
        client._login_err = gdata.service.CaptchaRequired
        client.captcha_url = 'a_captcha_url'
        client.captcha_token = 'token'
        self.get_input_called = False
        def _get_input(prompt):
            self.get_input_called = True
            delattr(client, '_login_err')
            return 'Foo'
        self.keyring = init_google_docs_keyring(client, input_getter=_get_input)
        self.keyring.client
        self.assertTrue(self.get_input_called, 'Should have got input')

    def test_retrieves_existing_keyring_with_and_without_bom(self):
        client = self._init_client()
        dummy_entries = dict(section1=dict(user1='pwd1'))
        no_utf8_bom_entries = self._encode_data(dummy_entries)
        client._request_response = dict(status=200, data=no_utf8_bom_entries)
        client._listfeed = self._init_listfeed()
        self._init_keyring(client)
        self.assertEqual(self.keyring.get_password('section1', 'user1'), 'pwd1')

        utf8_bom_entries = codecs.BOM_UTF8 + no_utf8_bom_entries
        client._request_response = dict(status=200, data=utf8_bom_entries)
        self._init_keyring(client)
        self.assertEqual(self.keyring.get_password('section1', 'user1'), 'pwd1')

    def test_handles_retrieve_failure(self):
        client = self._init_client()
        client._listfeed = self._init_listfeed()
        client._request_response = dict(status=400,
                                        reason='Data centre explosion')
        self._init_keyring(client)
        self.assertRaises(errors.InitError, self.keyring.get_password, 'any', 'thing')

    def test_handles_corrupt_retrieve(self):
        client = self._init_client()
        dummy_entries = dict(section1=dict(user1='pwd1'))
        client._request_response = dict(status=200, data='broken' + self._encode_data(dummy_entries))
        client._listfeed = self._init_listfeed()
        self._init_keyring(client)
        self.assertRaises(errors.InitError, self.keyring.get_password, 'any', 'thing')

    def test_no_create_if_requested(self):
        client = self._init_client()
        self.keyring = init_google_docs_keyring(client, can_create=False)
        self.assertRaises(errors.InitError, self.keyring.get_password, 'any', 'thing')

    def test_no_set_if_create_folder_fails_on_new_keyring(self):
        import gdata
        client = self._init_client()
        client._create_folder_err = gdata.service.RequestError
        self._init_keyring(client)
        self.assertEqual(self.keyring.get_password('service-a', 'user-A'), None,
                        'No password should be set in new keyring')
        self.assertRaises(errors.PasswordSetError, self.keyring.set_password,
                          'service-a', 'user-A', 'password-A')
        self.assertEqual(self.keyring.get_password('service-a', 'user-A'), None,
                        'No password should be set after write fail')

    def test_no_set_if_write_fails_on_new_keyring(self):
        import gdata
        client = self._init_client()
        client._upload_err = gdata.service.RequestError
        self._init_keyring(client)
        self.assertEqual(self.keyring.get_password('service-a', 'user-A'), None,
                        'No password should be set in new keyring')
        self.assertRaises(errors.PasswordSetError, self.keyring.set_password,
                          'service-a', 'user-A', 'password-A')
        self.assertEqual(self.keyring.get_password('service-a', 'user-A'), None,
                        'No password should be set after write fail')

    def test_no_set_if_write_fails_on_existing_keyring(self):
        import gdata
        client = self._init_client()
        dummy_entries = dict(sectionB=dict(user9='pwd9'))
        client._request_response = dict(status=200, data=self._encode_data(dummy_entries))
        client._put_err = gdata.service.RequestError
        client._listfeed = self._init_listfeed()
        self._init_keyring(client)
        self.assertEqual(self.keyring.get_password('sectionB', 'user9'), 'pwd9',
                        'Correct password should be set in existing keyring')
        self.assertRaises(errors.PasswordSetError, self.keyring.set_password,
                          'sectionB', 'user9', 'Not the same pwd')
        self.assertEqual(self.keyring.get_password('sectionB', 'user9'), 'pwd9',
                        'Password should be unchanged after write fail')

    def test_writes_correct_data_to_google_docs(self):
        client = self._init_client()
        dummy_entries = dict(sectionWriteChk=dict(userWriteChk='pwd'))
        client._request_response = dict(status=200, data=self._encode_data(dummy_entries))
        client._listfeed = self._init_listfeed()
        self._init_keyring(client)
        self.keyring.set_password('sectionWriteChk',
                                  'userWritechk',
                                  'new_pwd')
        self.assertIsNotNone(client._put_data, 'Should have written data')
        self.assertEquals(
            'new_pwd',
            client._put_data.get('sectionWriteChk').get('userWritechk'),
            'Did not write updated password!')

    def test_handles_write_conflict_on_different_service(self):
        import gdata
        client = self._init_client()
        dummy_entries = dict(sectionWriteConflictA=dict(
            userwriteConflictA='pwdwriteConflictA'))
        client._request_response = dict(status=200, data=self._encode_data(dummy_entries))
        client._put_err = [(gdata.service.RequestError,
                               {'status': '406',
                                'reason': 'Conflict'}),]
        client._listfeed = self._init_listfeed()
        self._init_keyring(client)
        self.assertEqual(
            self.keyring.get_password('sectionWriteConflictA',
                                      'userwriteConflictA'),
            'pwdwriteConflictA',
            'Correct password should be set in existing keyring')
        dummy_entries['diffSection'] = dict(foo='bar')
        client._request_response = dict(status=200, data=self._encode_data(dummy_entries))
        new_pwd = 'Not the same pwd'
        self.keyring.set_password('sectionWriteConflictA',
                                  'userwriteConflictA',
                                  new_pwd)

        self.assertEquals(self.keyring.get_password('sectionWriteConflictA',
                                                    'userwriteConflictA'),
                          new_pwd
        )
        self.assertEqual(1, client._put_count,
                         'Write not called after conflict resolution')

    def test_handles_write_conflict_on_same_service_and_username(self):
        import gdata
        client = self._init_client()
        dummy_entries = dict(sectionWriteConflictB=dict(
            userwriteConflictB='pwdwriteConflictB'))
        client._request_response = dict(status=200, data=self._encode_data(dummy_entries))
        client._put_err = (gdata.service.RequestError,
                               {'status': '406',
                                'reason': 'Conflict'})
        client._listfeed = self._init_listfeed()
        self._init_keyring(client)
        self.assertEqual(
            self.keyring.get_password('sectionWriteConflictB',
                                      'userwriteConflictB'),
            'pwdwriteConflictB',
            'Correct password should be set in existing keyring')
        conflicting_dummy_entries = dict(sectionWriteConflictB=dict(
            userwriteConflictB='pwdwriteConflictC'))
        client._request_response = dict(status=200, data=self._encode_data(conflicting_dummy_entries))
        self.assertRaises(errors.PasswordSetError, self.keyring.set_password,
                          'sectionWriteConflictB', 'userwriteConflictB', 'new_pwd')

    def test_handles_write_conflict_with_identical_change(self):
        import gdata
        client = self._init_client()
        dummy_entries = dict(sectionWriteConflictC=dict(
            userwriteConflictC='pwdwriteConflictC'))
        client._request_response = dict(status=200, data=self._encode_data(dummy_entries))
        client._put_err = [(gdata.service.RequestError,
                               {'status': '406',
                                 'reason': 'Conflict'}),]
        client._listfeed = self._init_listfeed()
        self._init_keyring(client)
        self.assertEqual(
            self.keyring.get_password('sectionWriteConflictC',
                                      'userwriteConflictC'),
            'pwdwriteConflictC',
            'Correct password should be set in existing keyring')
        new_pwd = 'Not the same pwd'
        conflicting_dummy_entries = dict(sectionWriteConflictC=dict(
            userwriteConflictC=new_pwd))
        client._request_response = dict(status=200, data=self._encode_data(conflicting_dummy_entries))
        self.keyring.set_password('sectionWriteConflictC',
                                  'userwriteConflictC',
                                  new_pwd)
        self.assertEquals(self.keyring.get_password('sectionWriteConflictC',
                                                    'userwriteConflictC'),
                          new_pwd
        )

    def test_handles_broken_google_put_when_non_owner_update_fails(self):
        """Google Docs has a bug when putting to a non-owner
           see  GoogleDocsKeyring._save_keyring()
        """
        import gdata
        client = self._init_client()
        dummy_entries = dict(sectionBrokenPut=dict(
            userBrokenPut='pwdBrokenPut'))
        client._request_response = dict(status=200, data=self._encode_data(dummy_entries))
        client._put_err = [(
            gdata.service.RequestError,
                { 'status': '400',
                  'body': 'Sorry, there was an error saving the file. Please try again.',
                  'reason': 'Bad Request'}),]
        client._listfeed = self._init_listfeed()
        self._init_keyring(client)
        new_pwd = 'newPwdBrokenPut'
        correct_read_entries = dict(sectionBrokenPut=dict(
            userBrokenPut='pwdBrokenPut'))
        client._request_response = dict(status=200,
                                        data=self._encode_data(correct_read_entries))
        self.assertRaises(errors.PasswordSetError, self.keyring.set_password,
                          'sectionBrokenPut', 'userBrokenPut', new_pwd)

    def test_handles_broken_google_put_when_non_owner_update(self):
        """Google Docs has a bug when putting to a non-owner
           see  GoogleDocsKeyring._save_keyring()
        """
        import gdata
        client = self._init_client()
        dummy_entries = dict(sectionBrokenPut=dict(
            userBrokenPut='pwdBrokenPut'))
        client._request_response = dict(status=200, data=self._encode_data(dummy_entries))
        client._put_err = [(
            gdata.service.RequestError,
                { 'status': '400',
                  'body': 'Sorry, there was an error saving the file. Please try again.',
                  'reason': 'Bad Request'}),]
        client._listfeed = self._init_listfeed()
        self._init_keyring(client)
        new_pwd = 'newPwdBrokenPut'
        correct_read_entries = dict(sectionBrokenPut=dict(
            userBrokenPut=new_pwd))
        client._request_response = dict(status=200,
                                        data=self._encode_data(correct_read_entries))
        self.keyring.set_password('sectionBrokenPut',
                                  'userBrokenPut',
                                  new_pwd)
        self.assertEquals(self.keyring.get_password('sectionBrokenPut',
                                                    'userBrokenPut'),
                          new_pwd)

    def test_uses_existing_folder(self):
        import gdata
        client = self._init_client()
        # should not happen
        client._create_folder_err = gdata.service.RequestError

        self._init_keyring(client)
        self.assertEqual(self.keyring.get_password('service-a', 'user-A'), None,
                         'No password should be set in new keyring')
        client._listfeed = self._init_listfeed()
        self.keyring.set_password('service-a', 'user-A', 'password-A')
        self.assertIsNotNone(client._upload_data, 'Should have written data')
        self.assertEqual(self.keyring.get_password('service-a', 'user-A'),
                         'password-A',
                         'Correct password should be set')



'''
Task Coach - Your friendly task manager
Copyright (C) 2014 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
import wx.lib.colourselect as csel
from wx.lib import sized_controls
from taskcoachlib.i18n import _
from taskcoachlib.thirdparty.wxScheduler import wxSCHEDULER_DAILY, \
    wxSCHEDULER_WEEKLY, wxSCHEDULER_MONTHLY, wxSCHEDULER_HORIZONTAL, \
    wxSCHEDULER_VERTICAL


class HierarchicalCalendarConfigDialog(sized_controls.SizedDialog):
    def __init__(self, settings, settingsSection, *args, **kwargs):
        self._settings = settings
        self._settingsSection = settingsSection
        super(HierarchicalCalendarConfigDialog, self).__init__(*args, **kwargs)
        pane = self.GetContentsPane()
        pane.SetSizerType('form')
        self.createInterior(pane)
        buttonSizer = self.CreateStdDialogButtonSizer(wx.OK|wx.CANCEL)
        self.SetButtonSizer(buttonSizer)
        self.Fit()
        buttonSizer.GetAffirmativeButton().Bind(wx.EVT_BUTTON, self.ok)

    def createInterior(self, pane):
        self.createHeaderEntry(pane)
        self.createPeriodEntry(pane)
        self.createLineEntry(pane)
        self.createColorEntry(pane)

    def createHeaderEntry(self, pane):
        label = wx.StaticText(pane, label=_('Headers'))
        label.SetSizerProps(valign='center')
        hdr = self._settings.getint(self._settingsSection, 'headerformat')
        panel = sized_controls.SizedPanel(pane)
        panel.SetSizerType('vertical')
        self._weekNumber = wx.CheckBox(panel, label=_('Week number'))
        self._weekNumber.SetValue(hdr & 1)
        self._weekNumber.SetSizerProps(valign='center')
        self._dates = wx.CheckBox(panel, label=_('Date'))
        self._dates.SetValue(hdr & 2)
        self._dates.SetSizerProps(valign='center')
        panel.SetSizerProps(valign='center')
        panel.Fit()

    def createPeriodEntry(self, pane):
        label = wx.StaticText(pane, label=_('Calendar span'))
        label.SetSizerProps(valign='center')
        periods = (_('Week'), _('Work week'), _('Month'))
        self._spanType = wx.Choice(pane, choices=periods)  # pylint: disable=W0201
        self._spanType.SetSizerProps(valign='center')
        self._spanType.SetSelection(self._settings.getint(self._settingsSection,
                                                          'calendarformat'))

    def createLineEntry(self, pane):
        label = wx.StaticText(pane,
                              label=_('Draw a line showing the current time'))
        label.SetSizerProps(valign='center')
        self._shownow = wx.CheckBox(pane) # pylint: disable=W0201
        self._shownow.SetSizerProps(valign='center')
        self._shownow.SetValue(self._settings.getboolean(self._settingsSection,
                                                         'drawnow'))

    def createColorEntry(self, pane):
        label = wx.StaticText(pane, label=_('Color used to highlight the current day'))
        label.SetSizerProps(valign='center')
        hcolor = self._settings.get(self._settingsSection, 'todaycolor')
        if not hcolor:
            # The highlight color is too dark
            color = wx.SystemSettings.GetColour( wx.SYS_COLOUR_HIGHLIGHT )
            color = wx.Colour(int((color.Red() + 255) / 2),
                              int((color.Green() + 255) / 2),
                              int((color.Blue() + 255) / 2))
        else:
            color = wx.Colour(*tuple(map(int, hcolor.split(','))))  # pylint: disable=W0141
        self._highlight = csel.ColourSelect(pane, size=(100, 20))  # pylint: disable=W0201
        label.SetSizerProps(valign='center')
        self._highlight.SetColour(color)

    def ok(self, event=None):  # pylint: disable=W0613
        settings, section = self._settings, self._settingsSection
        settings.set(section, 'calendarformat', str(self._spanType.GetSelection()))
        settings.set(section, 'drawnow', str(self._shownow.GetValue()))
        color = self._highlight.GetColour()
        settings.set(section, 'todaycolor',
                     '%d,%d,%d' % (color.Red(), color.Green(), color.Blue()))
        hdr = 0
        if self._weekNumber.GetValue():
            hdr |= 1
        if self._dates.GetValue():
            hdr |= 2
        settings.set(section, 'headerformat', str(hdr))
        self.EndModal(wx.ID_OK)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>
Copyright (C) 2008 Rob McMullen <rob.mcmullen@gmail.com>
Copyright (C) 2008 Carl Zmola <zmola@acm.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import widgets, operating_system
from taskcoachlib.domain import date
from taskcoachlib.gui import artprovider
from taskcoachlib.i18n import _
from taskcoachlib.thirdparty import combotreebox
import datetime
from wx.lib import newevent
import wx
import wx.combo

 
DateTimeEntryEvent, EVT_DATETIMEENTRY = newevent.NewEvent()


class DateTimeEntry(widgets.DateTimeCtrl):
    defaultDateTime = date.DateTime()

    def __init__(self, parent, settings, initialDateTime=defaultDateTime, 
                 readonly=False, noneAllowed=True, showSeconds=False, 
                 suggestedDateTime=None, showRelative=False, adjustEndOfDay=False, units=None, *args, **kwargs):
        starthour = settings.getint('view', 'efforthourstart')
        endhour = settings.getint('view', 'efforthourend')
        interval = settings.getint('view', 'effortminuteinterval')
        super(DateTimeEntry, self).__init__(parent, noneAllowed=noneAllowed, 
                                            starthour=starthour, 
                                            endhour=endhour, interval=interval, 
                                            showSeconds=showSeconds,
                                            showRelative=showRelative, adjustEndOfDay=adjustEndOfDay, units=units)
        if readonly:
            self.Disable()
        # First set the initial value and then set the callback so that the
        # callback is not triggered for the initial value
        if initialDateTime == date.DateTime() and suggestedDateTime:
            self.setSuggested(suggestedDateTime)
        else:
            self.SetValue(initialDateTime)
        self.setCallback(self.onDateTimeCtrlEdited)
        
    def SetValue(self, newValue=None):
        super(DateTimeEntry, self).SetValue(newValue or self.defaultDateTime)

    def setSuggested(self, suggestedDateTime):
        super(DateTimeEntry, self).SetValue(suggestedDateTime)
        super(DateTimeEntry, self).SetNone()
        
    def onDateTimeCtrlEdited(self, *args, **kwargs):  # pylint: disable=W0613
        wx.PostEvent(self, DateTimeEntryEvent())            
        

class TimeDeltaEntry(widgets.PanelWithBoxSizer):
    # We can't inherit from widgets.masked.TextCtrl because that class expects
    # GetValue to return a string and we want to return a TimeDelta.
    
    defaultTimeDelta = date.TimeDelta()

    def __init__(self, parent, timeDelta=defaultTimeDelta, readonly=False, 
                 *args, **kwargs):
        super(TimeDeltaEntry, self).__init__(parent, *args, **kwargs)
        hours, minutes, seconds = timeDelta.hoursMinutesSeconds()
        self._entry = widgets.masked.TimeDeltaCtrl(self, hours, minutes, 
                                                   seconds, readonly,
                                                   timeDelta < self.defaultTimeDelta)
        if readonly:
            self._entry.Disable()
        self.add(self._entry, flag=wx.EXPAND | wx.ALL, proportion=1)
        self.fit()

    def NavigateBook(self, event):
        self.GetParent().NavigateBook(not event.ShiftDown())
        return True

    def GetValue(self):
        return date.parseTimeDelta(self._entry.GetValue())
    
    def SetValue(self, newTimeDelta):
        hours, minutes, seconds = newTimeDelta.hoursMinutesSeconds()
        negative = newTimeDelta < self.defaultTimeDelta
        self._entry.set_value(hours, minutes, seconds, negative)
            
    def Bind(self, *args, **kwargs):  # pylint: disable=W0221
        self._entry.Bind(*args, **kwargs)


class AmountEntry(widgets.PanelWithBoxSizer):
    def __init__(self, parent, amount=0.0, readonly=False, *args, **kwargs):
        super(AmountEntry, self).__init__(parent, *args, **kwargs)
        self._entry = self.createEntry(amount)
        if readonly:
            self._entry.Disable()
        self.add(self._entry)
        self.fit()

    def createEntry(self, amount):
        return widgets.masked.AmountCtrl(self, amount)

    def NavigateBook(self, event):
        self.GetParent().NavigateBook(not event.ShiftDown())
        return True

    def GetValue(self):
        return self._entry.GetValue()

    def SetValue(self, value):
        self._entry.SetValue(value)

    def Bind(self, *args, **kwargs):  # pylint: disable=W0221
        self._entry.Bind(*args, **kwargs)


PercentageEntryEvent, EVT_PERCENTAGEENTRY = newevent.NewEvent()
        
        
class PercentageEntry(widgets.PanelWithBoxSizer):
    def __init__(self, parent, percentage=0, *args, **kwargs):
        kwargs['orientation'] = wx.HORIZONTAL
        super(PercentageEntry, self).__init__(parent, *args, **kwargs)
        self._entry = self._createSpinCtrl(percentage)
        self._slider = self._createSlider(percentage)
        self.add(self._entry, flag=wx.ALL, proportion=0)
        self.add((5, -1), flag=wx.ALL, proportion=0)
        self.add(self._slider, flag=wx.ALL, proportion=1)
        self.fit()
        
    def _createSlider(self, percentage):
        slider = wx.Slider(self, value=percentage, style=wx.SL_AUTOTICKS,
                          minValue=0, maxValue=100, size=(150, -1))
        slider.SetTickFreq(25)
        slider.Bind(wx.EVT_SCROLL, self.onSliderScroll)
        return slider
        
    def _createSpinCtrl(self, percentage):
        entry = widgets.SpinCtrl(self, value=percentage,
            min=0, max=100, size=(60 if operating_system.isMac() else 50, -1))
        for eventType in wx.EVT_SPINCTRL, wx.EVT_KILL_FOCUS:
            entry.Bind(eventType, self.onSpin)
        return entry

    def GetValue(self):
        return self._entry.GetValue()

    def SetValue(self, value):
        self._entry.SetValue(value)
        self._slider.SetValue(value)
        
    def onSliderScroll(self, event):  # pylint: disable=W0613
        self.syncControl(self._entry, self._slider)
            
    def onSpin(self, event):  # pylint: disable=W0613
        self.syncControl(self._slider, self._entry)
            
    def syncControl(self, controlToWrite, controlToRead):
        value = controlToRead.GetValue()
        # Prevent potential endless loop by checking that we really need to set
        # the value:
        if controlToWrite.GetValue() != value:
            controlToWrite.SetValue(value)
        wx.PostEvent(self, PercentageEntryEvent())


FontEntryEvent, EVT_FONTENTRY = newevent.NewEvent()
        
        
class FontEntry(widgets.PanelWithBoxSizer):
    def __init__(self, parent, currentFont, currentColor, *args, **kwargs):
        kwargs['orientation'] = wx.HORIZONTAL
        super(FontEntry, self).__init__(parent, *args, **kwargs)
        self._fontCheckBox = self._createCheckBox(currentFont)
        self._fontPicker = self._createFontPicker(currentFont, currentColor)
        self.add(self._fontCheckBox, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, 
                 proportion=0)
        self.add(self._fontPicker, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, 
                 proportion=1)
        self.fit()
        
    def _createCheckBox(self, currentFont):
        checkBox = wx.CheckBox(self, label=_('Use font:'))
        checkBox.SetValue(currentFont is not None)
        checkBox.Bind(wx.EVT_CHECKBOX, self.onChecked)
        return checkBox
    
    def _createFontPicker(self, currentFont, currentColor):
        defaultFont = wx.SystemSettings_GetFont(wx.SYS_DEFAULT_GUI_FONT)
        picker = widgets.FontPickerCtrl(self,
            font=currentFont or defaultFont, colour=currentColor)
        picker.Bind(wx.EVT_FONTPICKER_CHANGED, self.onFontPicked)
        return picker
    
    def onChecked(self, event):
        event.Skip()
        wx.PostEvent(self, FontEntryEvent())
        
    def onFontPicked(self, event):
        event.Skip()
        self._fontCheckBox.SetValue(True)
        wx.PostEvent(self, FontEntryEvent())
    
    def GetValue(self):
        return self._fontPicker.GetSelectedFont() if \
            self._fontCheckBox.IsChecked() else None
            
    def SetValue(self, newFont):
        checked = newFont is not None
        self._fontCheckBox.SetValue(checked)
        if checked:
            self._fontPicker.SetSelectedFont(newFont)

    def GetColor(self):
        return self._fontPicker.GetSelectedColour()
            
    def SetColor(self, newColor):
        self._fontPicker.SetSelectedColour(newColor)


ColorEntryEvent, EVT_COLORENTRY = newevent.NewEvent()


class ColorEntry(widgets.PanelWithBoxSizer):
    def __init__(self, parent, currentColor, defaultColor, *args, **kwargs):
        kwargs['orientation'] = wx.HORIZONTAL
        super(ColorEntry, self).__init__(parent, *args, **kwargs)
        self._colorCheckBox = self._createCheckBox(currentColor)
        self._colorPicker = self._createColorPicker(currentColor, defaultColor)
        self.add(self._colorCheckBox, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, 
                 proportion=0)
        self.add(self._colorPicker, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, 
                 proportion=1)
        self.fit()

    def _createCheckBox(self, currentColor):
        checkBox = wx.CheckBox(self, label=_('Use color:'))
        checkBox.SetValue(currentColor is not None)
        checkBox.Bind(wx.EVT_CHECKBOX, self.onChecked)
        return checkBox
    
    def _createColorPicker(self, currentColor, defaultColor):
        # wx.ColourPickerCtrl on Mac OS X expects a wx.Colour and fails on tuples
        # so convert the tuples to a wx.Colour:
        currentColor = wx.Colour(*currentColor) if currentColor else defaultColor  # pylint: disable=W0142
        picker = wx.ColourPickerCtrl(self, col=currentColor)
        picker.Bind(wx.EVT_COLOURPICKER_CHANGED, self.onColorPicked)
        return picker
    
    def onChecked(self, event):
        event.Skip()
        wx.PostEvent(self, ColorEntryEvent())
        
    def onColorPicked(self, event):
        event.Skip()
        self._colorCheckBox.SetValue(True)
        wx.PostEvent(self, ColorEntryEvent())

    def GetValue(self):
        return self._colorPicker.GetColour() if self._colorCheckBox.IsChecked() \
            else None
            
    def SetValue(self, newColor):
        checked = newColor is not None
        self._colorCheckBox.SetValue(checked)
        if checked:
            self._colorPicker.SetColour(newColor)


IconEntryEvent, EVT_ICONENTRY = newevent.NewEvent()


class IconEntry(wx.combo.BitmapComboBox):
    def __init__(self, parent, currentIcon, *args, **kwargs):
        kwargs['style'] = wx.CB_READONLY
        super(IconEntry, self).__init__(parent, *args, **kwargs)
        imageNames = sorted(artprovider.chooseableItemImages.keys())
        size = (16, 16)
        for imageName in imageNames:
            label = artprovider.chooseableItemImages[imageName]
            bitmap = wx.ArtProvider_GetBitmap(imageName, wx.ART_MENU, size)
            item = self.Append(label, bitmap)
            self.SetClientData(item, imageName)
        self.SetSelection(imageNames.index(currentIcon))
        self.Bind(wx.EVT_COMBOBOX, self.onIconPicked)
        
    def onIconPicked(self, event):
        event.Skip()
        wx.PostEvent(self, IconEntryEvent())

    def GetValue(self):
        return self.GetClientData(self.GetSelection())
    
    def SetValue(self, newValue):
        for index in range(self.GetCount()):
            if newValue == self.GetClientData(index):
                self.SetSelection(index)
                break


ChoiceEntryEvent, EVT_CHOICEENTRY = newevent.NewEvent()


class ChoiceEntry(wx.Choice):
    def __init__(self, parent, choices, currentChoiceValue, *args, **kwargs):
        super(ChoiceEntry, self).__init__(parent, *args, **kwargs)
        for choiceValue, choiceText in choices:
            self.Append(choiceText, choiceValue)
            if choiceValue == currentChoiceValue:
                self.SetSelection(self.GetCount() - 1)
        if self.GetSelection() == wx.NOT_FOUND:
            # Force a selection if necessary:
            self.SetSelection(0)
        self.Bind(wx.EVT_CHOICE, self.onChoice)

    def onChoice(self, event):
        event.Skip()
        wx.PostEvent(self, ChoiceEntryEvent())

    def GetValue(self):
        return self.GetClientData(self.GetSelection())

    def SetValue(self, newValue):
        for index in range(self.GetCount()):
            if newValue == self.GetClientData(index):
                self.SetSelection(index)
                break


TaskEntryEvent, EVT_TASKENTRY = newevent.NewEvent()


class TaskEntry(wx.Panel):
    ''' A ComboTreeBox with tasks. This class does not inherit from the
        ComboTreeBox widget, because that widget is created using a
        factory function. '''

    def __init__(self, parent, rootTasks, selectedTask):
        ''' Initialize the ComboTreeBox, add the root tasks recursively and
            set the selection. '''
        super(TaskEntry, self).__init__(parent)
        self._createInterior()
        self._addTasksRecursively(rootTasks)
        self.SetValue(selectedTask)

    def __getattr__(self, attr):
        ''' Delegate unknown attributes to the ComboTreeBox. This is needed
            since we cannot inherit from ComboTreeBox, but have to use
            delegation. '''
        return getattr(self._comboTreeBox, attr)

    def _createInterior(self):
        ''' Create the ComboTreebox widget. '''
        # pylint: disable=W0201
        self._comboTreeBox = combotreebox.ComboTreeBox(self,
            style=wx.CB_READONLY | wx.CB_SORT | wx.TAB_TRAVERSAL)
        self._comboTreeBox.Bind(wx.EVT_COMBOBOX, self.onTaskSelected)
        boxSizer = wx.BoxSizer()
        boxSizer.Add(self._comboTreeBox, flag=wx.EXPAND, proportion=1)
        self.SetSizerAndFit(boxSizer)
        
    def _addTasksRecursively(self, tasks, parentItem=None):
        ''' Add tasks to the ComboTreeBox and then recursively add their
            subtasks. '''
        for task in tasks:
            self._addTaskRecursively(task, parentItem)

    def _addTaskRecursively(self, task, parentItem=None):
        ''' Add a task to the ComboTreeBox and then recursively add its
            subtasks. '''
        if not task.isDeleted():
            item = self._comboTreeBox.Append(task.subject(), parent=parentItem)
            self._comboTreeBox.SetClientData(item, task)
            self._addTasksRecursively(task.children(), item)

    def onTaskSelected(self, event):  # pylint: disable=W0613
        wx.PostEvent(self, TaskEntryEvent())
        
    def SetValue(self, task):
        ''' Select the given task. '''
        self._comboTreeBox.SetClientDataSelection(task)

    def GetValue(self):
        ''' Return the selected task. '''
        selection = self._comboTreeBox.GetSelection()
        return self._comboTreeBox.GetClientData(selection)


RecurrenceEntryEvent, EVT_RECURRENCEENTRY = newevent.NewEvent()


class RecurrenceEntry(wx.Panel):
    horizontalSpace = (3, -1)
    verticalSpace = (-1, 3)

    def __init__(self, parent, recurrence, settings, *args, **kwargs):
        super(RecurrenceEntry, self).__init__(parent, *args, **kwargs)
        recurrenceFrequencyPanel = wx.Panel(self)
        self._recurrencePeriodEntry = wx.Choice(recurrenceFrequencyPanel, 
            choices=[_('None'), _('Daily'), _('Weekly'), _('Monthly'), 
                     _('Yearly')])        
        self._recurrencePeriodEntry.Bind(wx.EVT_CHOICE, self.onRecurrencePeriodEdited)
        self._recurrenceFrequencyEntry = widgets.SpinCtrl(recurrenceFrequencyPanel, 
                                                          size=(50, -1), 
                                                          value=1, min=1)
        self._recurrenceFrequencyEntry.Bind(wx.EVT_SPINCTRL, self.onRecurrenceEdited)
        self._recurrenceStaticText = wx.StaticText(recurrenceFrequencyPanel, 
                                                   label='reserve some space')
        self._recurrenceSameWeekdayCheckBox = wx.CheckBox(recurrenceFrequencyPanel, 
            label=_('keeping dates on the same weekday'))
        self._recurrenceSameWeekdayCheckBox.Bind(wx.EVT_CHECKBOX, 
                                                 self.onRecurrenceEdited)
        panelSizer = wx.BoxSizer(wx.HORIZONTAL)
        panelSizer.Add(self._recurrencePeriodEntry, 
                       flag=wx.ALIGN_CENTER_VERTICAL)
        panelSizer.Add(self.horizontalSpace)
        panelSizer.Add(wx.StaticText(recurrenceFrequencyPanel, 
                                     label=_(', every')), 
                       flag=wx.ALIGN_CENTER_VERTICAL)
        panelSizer.Add(self.horizontalSpace)
        panelSizer.Add(self._recurrenceFrequencyEntry, 
                       flag=wx.ALIGN_CENTER_VERTICAL)
        panelSizer.Add(self.horizontalSpace)
        panelSizer.Add(self._recurrenceStaticText, 
                       flag=wx.ALIGN_CENTER_VERTICAL)
        panelSizer.Add(self.horizontalSpace)
        panelSizer.Add(self._recurrenceSameWeekdayCheckBox, proportion=1, 
                       flag=wx.ALIGN_CENTER_VERTICAL | wx.EXPAND)
        recurrenceFrequencyPanel.SetSizerAndFit(panelSizer)
        self._recurrenceSizer = panelSizer

        maxPanel = wx.Panel(self)
        panelSizer = wx.BoxSizer(wx.HORIZONTAL)
        
        self._maxRecurrenceCheckBox = wx.CheckBox(maxPanel)
        self._maxRecurrenceCheckBox.Bind(wx.EVT_CHECKBOX, 
                                         self.onMaxRecurrenceChecked)
        self._maxRecurrenceCountEntry = widgets.SpinCtrl(maxPanel, 
                                                         size=(50, -1), 
                                                         value=1, min=1)
        self._maxRecurrenceCountEntry.Bind(wx.EVT_SPINCTRL, 
                                           self.onRecurrenceEdited)
        panelSizer.Add(self._maxRecurrenceCheckBox, 
                       flag=wx.ALIGN_CENTER_VERTICAL)
        panelSizer.Add(self.horizontalSpace)
        panelSizer.Add(wx.StaticText(maxPanel, label=_('Stop after')),
                       flag=wx.ALIGN_CENTER_VERTICAL)
        panelSizer.Add(self.horizontalSpace)
        panelSizer.Add(self._maxRecurrenceCountEntry,
                       flag=wx.ALIGN_CENTER_VERTICAL)
        panelSizer.Add(self.horizontalSpace)
        panelSizer.Add(wx.StaticText(maxPanel, label=_('recurrences')),
                       flag=wx.ALIGN_CENTER_VERTICAL)
        maxPanel.SetSizerAndFit(panelSizer)
        
        schedulePanel = wx.Panel(self)
        panelSizer = wx.BoxSizer(wx.HORIZONTAL)
        label = wx.StaticText(schedulePanel, 
                              label=_('Schedule each next recurrence based on'))
        panelSizer.Add(label, flag=wx.ALIGN_CENTER_VERTICAL)
        panelSizer.Add((3, -1))
        self._scheduleChoice = wx.Choice(schedulePanel,
            choices=[_('previous planned start and/or due date'),
                     _('last completion date')])
        self._scheduleChoice.Bind(wx.EVT_CHOICE, self.onRecurrenceEdited)
        if operating_system.isMac():
            # On Mac OS X, the wx.Choice gets too little vertical space by 
            # default
            size = self._scheduleChoice.GetSizeTuple()
            self._scheduleChoice.SetMinSize((size[0], size[1] + 1))
        panelSizer.Add(self._scheduleChoice, flag=wx.ALIGN_CENTER_VERTICAL)
        schedulePanel.SetSizerAndFit(panelSizer)

        stopPanel = wx.Panel(self)
        panelSizer = wx.BoxSizer(wx.HORIZONTAL)

        self._stopDateTimeCheckBox = wx.CheckBox(stopPanel)
        self._stopDateTimeCheckBox.Bind(wx.EVT_CHECKBOX, 
                                        self.onRecurrenceStopDateTimeChecked)
        # Since None is not allowed, we need an initial value...
        self._recurrenceStopDateTimeEntry = DateTimeEntry(stopPanel, settings,
                                                          noneAllowed=False,
            initialDateTime=datetime.datetime.combine(date.LastDayOfCurrentMonth(),
                                                      datetime.time(0, 0, 0)))
        self._recurrenceStopDateTimeEntry.Bind(EVT_DATETIMEENTRY,
                                               self.onRecurrenceEdited)
        panelSizer.Add(self._stopDateTimeCheckBox, 
                       flag=wx.ALIGN_CENTER_VERTICAL)
        panelSizer.Add(self.horizontalSpace)
        panelSizer.Add(wx.StaticText(stopPanel, label=_('Stop after')),
                       flag=wx.ALIGN_CENTER_VERTICAL)
        panelSizer.Add(self.horizontalSpace)
        panelSizer.Add(self._recurrenceStopDateTimeEntry,
                       flag=wx.ALIGN_CENTER_VERTICAL)
        panelSizer.Add(self.horizontalSpace)
        stopPanel.SetSizerAndFit(panelSizer)
        
        panelSizer = wx.BoxSizer(wx.VERTICAL)
        panelSizer.Add(recurrenceFrequencyPanel)
        panelSizer.Add(self.verticalSpace)
        panelSizer.Add(schedulePanel)
        panelSizer.Add(self.verticalSpace)
        panelSizer.Add(maxPanel)
        panelSizer.Add(stopPanel)
        self.SetSizerAndFit(panelSizer)
        self.SetValue(recurrence)

    def updateRecurrenceLabel(self):
        recurrenceDict = {0: _('period,'), 1: _('day(s),'), 2: _('week(s),'),
                          3: _('month(s),'), 4: _('year(s),')}
        recurrenceLabel = recurrenceDict[self._recurrencePeriodEntry.Selection]
        self._recurrenceStaticText.SetLabel(recurrenceLabel)
        self._recurrenceSameWeekdayCheckBox.Enable(self._recurrencePeriodEntry.Selection in (3, 4))
        self._recurrenceSizer.Layout()

    def onRecurrencePeriodEdited(self, event):
        recurrenceOn = event.String != _('None')
        self._maxRecurrenceCheckBox.Enable(recurrenceOn)
        self._stopDateTimeCheckBox.Enable(recurrenceOn)
        self._recurrenceFrequencyEntry.Enable(recurrenceOn)
        self._scheduleChoice.Enable(recurrenceOn)
        self._maxRecurrenceCountEntry.Enable(recurrenceOn and \
            self._maxRecurrenceCheckBox.IsChecked())
        self._recurrenceStopDateTimeEntry.Enable(recurrenceOn and \
            self._stopDateTimeCheckBox.IsChecked())
        self.updateRecurrenceLabel()
        self.onRecurrenceEdited()

    def onMaxRecurrenceChecked(self, event):
        maxRecurrenceOn = event.IsChecked()
        self._maxRecurrenceCountEntry.Enable(maxRecurrenceOn)
        self.onRecurrenceEdited()

    def onRecurrenceStopDateTimeChecked(self, event):
        stopRecurrenceOn = event.IsChecked()
        self._recurrenceStopDateTimeEntry.Enable(stopRecurrenceOn)
        self.onRecurrenceEdited()

    def onRecurrenceEdited(self, event=None):  # pylint: disable=W0613
        wx.PostEvent(self, RecurrenceEntryEvent())
        
    def SetValue(self, recurrence):
        index = {'': 0, 'daily': 1, 'weekly': 2, 'monthly': 3, 
                 'yearly': 4}[recurrence.unit]
        self._recurrencePeriodEntry.Selection = index
        self._maxRecurrenceCheckBox.Enable(bool(recurrence))
        self._maxRecurrenceCheckBox.SetValue(recurrence.max > 0)
        self._maxRecurrenceCountEntry.Enable(recurrence.max > 0)
        if recurrence.max > 0:
            self._maxRecurrenceCountEntry.Value = recurrence.max
        self._recurrenceFrequencyEntry.Enable(bool(recurrence))
        self._recurrenceFrequencyEntry.Value = recurrence.amount
        self._recurrenceSameWeekdayCheckBox.Value = recurrence.sameWeekday \
            if recurrence.unit in ('monthly', 'yearly') else False
        self._scheduleChoice.Selection = 1 if recurrence.recurBasedOnCompletion else 0
        self._scheduleChoice.Enable(bool(recurrence))
        self._stopDateTimeCheckBox.Enable(bool(recurrence))
        has_stop_datetime = recurrence.stop_datetime != date.DateTime()
        self._stopDateTimeCheckBox.SetValue(has_stop_datetime)
        self._recurrenceStopDateTimeEntry.Enable(has_stop_datetime)
        if has_stop_datetime:
            self._recurrenceStopDateTimeEntry.SetValue(recurrence.stop_datetime)
        self.updateRecurrenceLabel()

    def GetValue(self):
        recurrenceDict = {0: '', 1: 'daily', 2: 'weekly', 3: 'monthly', 
                          4: 'yearly'}
        kwargs = dict(unit=recurrenceDict[self._recurrencePeriodEntry.Selection])
        if self._maxRecurrenceCheckBox.IsChecked():
            kwargs['maximum'] = self._maxRecurrenceCountEntry.Value
        kwargs['amount'] = self._recurrenceFrequencyEntry.Value
        kwargs['sameWeekday'] = self._recurrenceSameWeekdayCheckBox.IsChecked()
        kwargs['recurBasedOnCompletion'] = bool(self._scheduleChoice.Selection)
        if self._stopDateTimeCheckBox.IsChecked():
            kwargs['stop_datetime'] = self._recurrenceStopDateTimeEntry.GetValue()
        return date.Recurrence(**kwargs)  # pylint: disable=W0142




'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import time, os
import test, mock
from taskcoachlib import persistence, config
from taskcoachlib.domain import task, category, note
from taskcoachlib.syncml.config import createDefaultSyncConfig


class PerformanceTest(test.TestCase):
    def createTestFile(self):
        task.Task.settings = config.Settings(load=False)
        taskList = task.TaskList([task.Task('test') for _ in range(self.nrTasks)])
        taskfile = file(self.taskfilename, 'w')
        taskWriter = persistence.XMLWriter(taskfile)
        taskWriter.write(taskList, category.CategoryList(), note.NoteContainer(),
                         createDefaultSyncConfig('fake'), 'fake')
        taskfile.close()

    def setUp(self):
        self.nrTasks = 100
        self.taskfilename = 'performanceTest.tsk'
        self.createTestFile()

    def tearDown(self):
        os.remove(self.taskfilename)
        super(PerformanceTest, self).tearDown()

    def testRead(self):
        mockApp = mock.App()
        start = time.time()
        mockApp.iocontroller.open(self.taskfilename)
        end = time.time()
        self.assertEqual(self.nrTasks, len(mockApp.taskFile.tasks()))
        self.failUnless(end-start < self.nrTasks/10)
        mockApp.quitApplication()



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx


class FontMixer(object):
    @classmethod
    def mix(class_, *fonts):
        fonts = [font for font in fonts if font]
        if not fonts:
            return None
        elif len(fonts) == 1:
            return fonts[0]
        pointSize = class_.mixFontSizes(*fonts)
        family = class_.mixFontFamilies(*fonts)
        weight = class_.mixFontWeights(*fonts)
        style = class_.mixFontStyles(*fonts)
        underlined = class_.mixFontUnderlining(*fonts)
        return wx.Font(pointSize, family, style, weight, underline=underlined)
    
    @staticmethod
    def mixFontSizes(*fonts):
        size = 0
        for font in fonts:
            size += font.GetPointSize()
        return size/len(fonts)

    allFamilies = (wx.FONTFAMILY_SWISS, wx.FONTFAMILY_DECORATIVE,
                   wx.FONTFAMILY_ROMAN, wx.FONTFAMILY_SCRIPT, 
                   wx.FONTFAMILY_MODERN, wx.FONTFAMILY_TELETYPE)
     
    @classmethod
    def mixFontFamilies(class_, *fonts):
        families = [font.GetFamily() for font in fonts]
        counts = dict()
        for family in class_.allFamilies:
            counts[family] = families.count(family)    
        for family in class_.allFamilies:
            countsCopy = counts.copy()
            familyCount = countsCopy.pop(family)
            if familyCount > max(countsCopy.values()):
                return family
        return wx.FONTFAMILY_DEFAULT
    
    @staticmethod
    def mixFontWeights(*fonts):
        weights = [font.GetWeight() for font in fonts]
        countLight = weights.count(wx.FONTWEIGHT_LIGHT)
        countBold = weights.count(wx.FONTWEIGHT_BOLD)
        if countBold > countLight:
            return wx.FONTWEIGHT_BOLD
        if countLight > countBold:
            return wx.FONTWEIGHT_LIGHT
        return wx.FONTWEIGHT_NORMAL
    
    @staticmethod
    def mixFontStyles(*fonts):
        ''' If any of the fonts is italic, return italic else normal. We
            ignore slant style since a font created with the 
            wx.FONTSTYLE_SLANT style returns wx.FONTSTYLE_ITALIC as its 
            style. '''
        anyItalic = wx.FONTSTYLE_ITALIC in [font.GetStyle() for font in fonts] 
        return wx.FONTSTYLE_ITALIC if anyItalic else wx.FONTSTYLE_NORMAL
    
    @staticmethod
    def mixFontUnderlining(*fonts):
        return any(font.GetUnderlined() for font in fonts)
    



'''
Task Coach - Your friendly task manager
Copyright (C) 2014 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import operating_system
import re

class StrftimeFix(object):
    """Mixin class to fix strftime() so that it works with years < 1900"""

    def strftime(self, *args):
        if self.year >= 1900:
            return operating_system.decodeSystemString(super(StrftimeFix, self).strftime(*args))
        result = self.replace(year=self.year + 1900).strftime(*args)
        return re.sub(str(self.year + 1900), str(self.year), result)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import patterns, config
from taskcoachlib.domain import task, effort, date
import test


class EffortAggregatorTestCase(test.TestCase):
    aggregation = 'One of: day, week, or month (override in subclass)'
    
    def setUp(self):
        task.Task.settings = config.Settings(load=False)
        self.taskList = task.TaskList()
        self.effortAggregator = effort.EffortAggregator(self.taskList, 
            aggregation=self.aggregation)
        patterns.Publisher().registerObserver(self.onEvent, 
            eventType=self.effortAggregator.addItemEventType(),
            eventSource=self.effortAggregator)
        patterns.Publisher().registerObserver(self.onEvent,
            eventType=self.effortAggregator.removeItemEventType(),
            eventSource=self.effortAggregator)
        self.task1 = task.Task(subject='task 1')
        self.task2 = task.Task(subject='task 2')
        self.task3 = task.Task(subject='child')
        self.task1.addChild(self.task3)
        self.task3.setParent(self.task1)
        self.effort1period1a = effort.Effort(self.task1, 
            date.DateTime(2004, 1, 1, 11, 0, 0), 
            date.DateTime(2004, 1, 1, 12, 0, 0))
        self.effort2period1a = effort.Effort(self.task2, 
            date.DateTime(2004, 1, 1, 11, 0, 0), 
            date.DateTime(2004, 1, 1, 12, 0, 0))
        self.effort1period1b = effort.Effort(self.task1, 
            date.DateTime(2004, 1, 1, 13, 0, 0), 
            date.DateTime(2004, 1, 1, 14, 0, 0))
        self.effort2period1b = effort.Effort(self.task2, 
            date.DateTime(2004, 1, 1, 13, 0, 0), 
            date.DateTime(2004, 1, 1, 14, 0, 0))
        self.effort1period2 = effort.Effort(self.task1, 
            date.DateTime(2004, 2, 2, 13, 0, 0), 
            date.DateTime(2004, 2, 2, 14, 0, 0))
        self.effort1period3 = effort.Effort(self.task1,
            date.DateTime(2004, 1, 1, 10, 0, 0), 
            date.DateTime(2005, 1, 1, 10, 0, 0))
        self.effort3period1a = effort.Effort(self.task3, 
            date.DateTime(2004, 1, 1, 14, 0, 0), 
            date.DateTime(2004, 1, 1, 15, 0, 0))
        self.events = []

    def onEvent(self, event):
        self.events.append(event)

                
class CommonTestsMixin(object):
    def testEmptyTaskList(self):
        self.assertEqual(0, len(self.effortAggregator))
        
    def testAddTaskWithoutEffort(self):
        self.taskList.append(self.task1)
        self.assertEqual(0, len(self.effortAggregator))
        
    def testAddTaskWithEffort(self):
        self.task1.addEffort(self.effort1period1a)
        self.taskList.append(self.task1)
        self.assertEqual(2, len(self.effortAggregator))

    def testAddEffort(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(effort.Effort(self.task1))
        self.assertEqual(2, len(self.effortAggregator))
        
    def testAddTwoEffortsOnSameDay(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(self.effort1period1a)
        self.task1.addEffort(self.effort1period1b)
        self.assertEqual(2, len(self.effortAggregator))

    def testAddTaskWithTwoEffortsOnSameDay(self):
        self.task1.addEffort(self.effort1period1a)
        self.task1.addEffort(self.effort1period1b)
        self.taskList.append(self.task1)
        self.assertEqual(2, len(self.effortAggregator))
        
    def testAddTaskWithTwoEffortsOnSameDayAndCheckTotalEffort(self):
        self.task1.addEffort(self.effort1period1a)
        self.task1.addEffort(self.effort1period1b)
        self.taskList.append(self.task1)
        expectedDuration = self.effort1period1a.duration() + \
                           self.effort1period1b.duration()
        self.assertEqual(expectedDuration, 
                         list(self.effortAggregator)[0].duration())
        
    def testAddTwoEffortsInDifferentPeriods(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(self.effort1period1a)
        self.task1.addEffort(self.effort1period2)
        self.assertEqual(4, len(self.effortAggregator))

    def testAddTwoEffortsOnTheSameDayToTwoDifferentTasks(self):
        self.taskList.extend([self.task1, self.task2])
        self.task1.addEffort(self.effort1period1a)
        self.task2.addEffort(self.effort2period1a)
        self.assertEqual(3, len(self.effortAggregator))

    def testAddEffortToChild(self):
        self.taskList.extend([self.task1, self.task2])
        self.task1.addChild(self.task2)
        self.task2.addEffort(self.effort2period1a)
        self.assertEqual(3, len(self.effortAggregator))

    def testAddChildWithEffort(self):
        self.taskList.extend([self.task1, self.task2])
        self.task2.addEffort(self.effort2period1a)
        self.task1.addChild(self.task2)
        self.assertEqual(3, len(self.effortAggregator))

    def testAddParentAndChildWithEffortToTaskList(self):
        self.task3.addEffort(self.effort3period1a)
        self.taskList.append(self.task1)
        self.assertEqual(3, len(self.effortAggregator))

    def testAddEffortToGrandChild(self):
        self.taskList.extend([self.task1, self.task2])
        self.task3.addChild(self.task2)
        self.task2.addEffort(self.effort2period1a)
        self.assertEqual(4, len(self.effortAggregator))

    def testAddGrandChildWithEffort(self):
        self.taskList.extend([self.task1, self.task2])
        self.task2.addEffort(self.effort2period1a)
        self.task3.addChild(self.task2)
        self.assertEqual(4, len(self.effortAggregator))

    def testRemoveChildWithEffortFromParent(self):
        self.taskList.extend([self.task1, self.task2])
        self.task1.addChild(self.task2)
        self.task2.addEffort(self.effort2period1a)
        self.task2.setParent(None)
        self.task1.removeChild(self.task2)
        self.assertEqual(2, len(self.effortAggregator))

    def testRemoveEffort(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(self.effort1period1a)
        self.task1.removeEffort(self.effort1period1a)
        self.assertEqual(0, len(self.effortAggregator))

    def testRemoveOneOfTwoEfforts(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(self.effort1period1a)
        self.task1.addEffort(self.effort1period1b)
        self.task1.removeEffort(self.effort1period1a)
        self.assertEqual(2, len(self.effortAggregator))

    def testRemoveOneOfTwoEffortsOfDifferentTasks(self):
        self.taskList.extend([self.task1, self.task2])
        self.task1.addEffort(self.effort1period1a)
        self.task2.addEffort(self.effort2period1a)
        self.task1.removeEffort(self.effort1period1a)
        self.assertEqual(2, len(self.effortAggregator))

    def testRemoveTwoOfTwoEfforts(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(self.effort1period1a)
        self.task1.addEffort(self.effort1period1b)
        self.task1.removeEffort(self.effort1period1a)
        self.task1.removeEffort(self.effort1period1b)
        self.assertEqual(0, len(self.effortAggregator))

    def testRemoveEffortFromChild(self):
        self.taskList.extend([self.task1, self.task2])
        self.task2.addEffort(self.effort2period1a)
        self.task1.addChild(self.task2)
        self.task2.removeEffort(self.effort2period1a)
        self.assertEqual(0, len(self.effortAggregator))

    def testRemoveTasks(self):
        self.taskList.extend([self.task1, self.task3])
        self.task3.addEffort(self.effort3period1a)
        self.taskList.removeItems([self.task1, self.task3])
        self.assertEqual(0, len(self.effortAggregator))

    def testRemoveTasksWithOverlappingEffort(self):
        self.taskList.extend([self.task1, self.task3])
        self.task3.addEffort(self.effort3period1a)
        self.task1.addEffort(self.effort1period1a)
        self.taskList.removeItems([self.task1, self.task3])
        self.assertEqual(0, len(self.effortAggregator))

    def testRemoveAllTasks(self):
        self.taskList.extend([self.task1, self.task2, self.task3])
        self.task3.addEffort(self.effort3period1a)
        self.taskList.removeItems([self.task1, self.task2, self.task3])
        self.assertEqual(0, len(self.effortAggregator))
 
    def testRemoveChildTask(self):
        self.taskList.extend([self.task1])
        self.task3.addEffort(self.effort3period1a)
        self.taskList.removeItems([self.task3])
        self.assertEqual(0, len(self.effortAggregator))
 
    def testChangeStart(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(self.effort1period1a)
        self.effort1period1a.setStart(date.DateTime.now())
        self.assertEqual(2, len(self.effortAggregator))

    def testChangeStartOfOneOfTwoEfforts(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(self.effort1period1a)
        self.task1.addEffort(self.effort1period1b)
        self.effort1period1a.setStart(date.DateTime.now())
        self.assertEqual(4, len(self.effortAggregator))

    def testChangeStart_WithinPeriod(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(self.effort1period1a)
        self.effort1period1a.setStart(self.effort1period1a.getStart() + \
            date.ONE_SECOND)
        self.assertEqual(2, len(self.effortAggregator))

    def testChangeStopDoesNotAffectPeriod(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(self.effort1period1a)
        composite = list(self.effortAggregator)[0]
        start = composite.getStart()
        self.effort1period1a.setStop(date.DateTime.now())
        self.assertEqual(start, composite.getStart())

    def testChangeStartOfOneOfTwoEffortsToOneYearLater(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(self.effort1period1a)
        self.task1.addEffort(self.effort1period1b)
        self.effort1period1a.setStart(date.DateTime(2005, 1, 1, 11, 0, 0))
        self.assertEqual(4, len(self.effortAggregator))

    def testNotification_Add(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(self.effort1period1a)
        self.assertEqual(1, len(self.events))
        self.failUnless(self.events[0].value() in self.effortAggregator)

    def testNotification_Remove(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(self.effort1period1a)
        self.task1.removeEffort(self.effort1period1a)
        self.assertEqual(3, len(self.events))

    def testCreateWithInitialEffort(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(self.effort1period1a)
        aggregator = effort.EffortAggregator(self.taskList, 
            aggregation=self.aggregation)
        self.assertEqual(2, len(aggregator))

    def testLongEffortIsStillOneCompositeEffort(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(self.effort1period3)
        self.assertEqual(2, len(self.effortAggregator))

    def testChangeTask(self):
        self.taskList.extend([self.task1, self.task2])
        self.task1.addEffort(self.effort1period1a)
        self.effort1period1a.setTask(self.task2)
        self.assertEqual(2, len(self.effortAggregator))
        self.failUnless(self.task2 in [item.task() for item in self.effortAggregator])

    def testChangeTaskOfChildEffort(self):
        self.taskList.extend([self.task1, self.task2])
        self.task3.addEffort(self.effort3period1a)
        self.effort3period1a.setTask(self.task2)
        self.assertEqual(2, len(self.effortAggregator))
        self.failUnless(self.task2 in [item.task() for item in self.effortAggregator])

    def testRemoveTaskAfterChangeTaskOfEffort(self):
        self.taskList.extend([self.task1, self.task2])
        self.task1.addEffort(self.effort1period1a)
        self.effort1period1a.setTask(self.task2)
        self.taskList.remove(self.task1)
        self.assertEqual(2, len(self.effortAggregator))
        self.failUnless(self.task2 in [item.task() for item in self.effortAggregator])

    def testRemoveAndAddEffortToSamePeriod(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(self.effort1period1a)
        self.task1.removeEffort(self.effort1period1a)
        self.task1.addEffort(self.effort1period1a)
        self.assertEqual(2, len(self.effortAggregator))
        self.assertEqual(self.effort1period1a, list(self.effortAggregator)[0][0])

    def testMaxDateTime(self):
        self.assertEqual(None, self.effortAggregator.maxDateTime())

    def testMaxDateTime_OneEffort(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(self.effort1period1a)
        self.assertEqual(self.effort1period1a.getStop(), 
            self.effortAggregator.maxDateTime())

    def testMaxDateTime_OneTrackingEffort(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(effort.Effort(self.task1))
        self.assertEqual(None, self.effortAggregator.maxDateTime())

    def testMaxDateTime_TwoEfforts(self):
        self.taskList.append(self.task1)
        self.task1.addEffort(self.effort1period1a)
        now = date.DateTime.now()
        self.task1.addEffort(effort.Effort(self.task1, 
            self.effort1period1a.getStart(), now))
        self.assertEqual(now, self.effortAggregator.maxDateTime())
   
    def testNrTracking(self):
        self.assertEqual(0, self.effortAggregator.nrBeingTracked())

    def testOriginalLength(self):
        self.assertEqual(0, self.effortAggregator.originalLength())


class EffortPerDayTest(EffortAggregatorTestCase, CommonTestsMixin):
    aggregation = 'day'


class EffortPerWeekTest(EffortAggregatorTestCase, CommonTestsMixin):
    aggregation = 'week'


class EffortPerMonthTest(EffortAggregatorTestCase, CommonTestsMixin):
    aggregation = 'month'

        
class MultipleAggregatorsTest(test.TestCase):
    def setUp(self):
        self.taskList = task.TaskList()
        self.effortPerDay = effort.EffortSorter(effort.EffortAggregator(self.taskList, aggregation='day'))
        self.effortPerWeek = effort.EffortSorter(effort.EffortAggregator(self.taskList, aggregation='week'))
        
    def testDeleteEffort_StartOfBothPeriods(self):
        aTask = task.Task()
        self.taskList.append(aTask)
        # Make sure the start of the day and week are the same, 
        # in other words, use a Monday
        anEffort = effort.Effort(aTask, date.DateTime(2006, 8, 28), 
                                 date.DateTime(2006, 8, 29))
        aTask.addEffort(anEffort)
        aTask.removeEffort(anEffort)
        self.failIf(self.effortPerDay)
